import numpy as np
from zounds.nputil import sliding_window, windowed
from zounds.util import tuplify
from dimensions import IdentityDimension


class CustomSlice(object):
    def __init__(self):
        super(CustomSlice, self).__init__()


class ArrayWithUnits(np.ndarray):
    """
    `ArrayWithUnits` is an :class:`numpy.ndarray` subclass that allows for
    indexing by more semantically meaningful slices.

    It supports most methods on :class:`numpy.ndarray`, and makes a best-effort
    to maintain meaningful dimensions throughout those operations.

    Args:
        arr (ndarray): The :class:`numpy.ndarray` instance containing the raw
            data for this instance
        dimensions (list or tuple): list or tuple of :class:`Dimension`-derived
            classes

    Raises:
        ValueError: when `arr.ndim` and `len(dimensions)` do not match

    Examples:
        >>> from zounds import ArrayWithUnits, TimeDimension, Seconds, TimeSlice
        >>> import numpy as np
        >>> data = np.zeros(100)
        >>> awu = ArrayWithUnits(data, [TimeDimension(Seconds(1))])
        >>> sliced = awu[TimeSlice(Seconds(10))]
        >>> sliced.shape
        (10,)

    See Also:
        :class:`IdentityDimension`
        :class:`~zounds.timeseries.TimeDimension`
        :class:`~zounds.spectral.FrequencyDimension`
    """

    def __new__(cls, arr, dimensions):
        if arr.ndim != len(dimensions):
            raise ValueError(
                'arr.ndim and len(dimensions) must match.  '
                'They were {arr.shape} and {dimensions}'.format(**locals()))

        obj = np.asarray(arr).view(cls)
        obj.dimensions = tuple(map(
            lambda d: IdentityDimension() if d is None else d.copy(),
            dimensions))

        for dim, size in zip(obj.dimensions, obj.shape):
            try:
                dim.size = size
            except AttributeError:
                pass
            try:
                dim.validate(size)
            except AttributeError:
                pass

        return obj

    def kwargs(self):
        return self.__dict__

    def concatenate(self, other):
        if self.dimensions == other.dimensions:
            return self.from_example(np.concatenate([self, other]), self)
        else:
            raise ValueError('All dimensions must match to concatenate')

    def reshape(self, shape, order='C'):
        if shape == filter(lambda x: x > 1, self.shape):
            return self.squeeze()

        raw = np.asarray(self)
        return np.reshape(raw, shape, order)

    def squeeze(self):
        zipped = filter(lambda x: x[0] > 1, zip(self.shape, self.dimensions))
        return ArrayWithUnits(
            np.reshape(self, [s for s, _ in zipped]),
            [d for _, d in zipped])

    @classmethod
    def concat(cls, arrs, axis=0):
        for arr in arrs[1:]:
            if arr.dimensions != arrs[0].dimensions:
                raise ValueError('All dimensions must match')
        return cls.from_example(np.concatenate(arrs, axis=axis), arrs[0])

    @classmethod
    def from_example(cls, data, example):
        """
        Produce a new :class:`ArrayWithUnits` instance given some raw data and
        an example instance that has the desired dimensions
        """
        return ArrayWithUnits(data, example.dimensions)

    @classmethod
    def zeros(cls, example):
        return cls.from_example(
            np.zeros(example.shape, dtype=example.dtype), example)

    def _apply_reduction_to_dimensions(self, result, axis, keepdims):
        if axis is None:
            return result

        ndims = len(self.dimensions)
        reduced_axes = set([ndims + a if a < 0 else a for a in tuplify(axis)])
        all_axes = set(range(ndims))

        if keepdims:
            new_dims = [
                IdentityDimension() if i in reduced_axes else dim
                for i, dim in enumerate(self.dimensions)]
        else:
            remaining_axes = sorted(all_axes - reduced_axes)
            new_dims = [self.dimensions[i] for i in remaining_axes]

        return ArrayWithUnits(result, new_dims)

    def sum(self, axis=None, dtype=None, keepdims=False, **kwargs):
        result = super(ArrayWithUnits, self).sum(
            axis, dtype, keepdims=keepdims, **kwargs)
        return self._apply_reduction_to_dimensions(result, axis, keepdims)

    def max(self, axis=None, out=None, keepdims=False):
        result = super(ArrayWithUnits, self).max(
            axis=axis, out=out, keepdims=keepdims)
        return self._apply_reduction_to_dimensions(result, axis, keepdims)

    def dot(self, b):
        result = super(ArrayWithUnits, self).dot(b)
        return self.__class__(result, self.dimensions[:result.ndim])

    def packbits(self, axis=None):
        arr = np.packbits(self, axis=axis)
        return ArrayWithUnits(arr, self.dimensions)

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.dimensions = getattr(obj, 'dimensions', None)

    def __array_wrap__(self, obj, context=None):
        if len(self.dimensions) != obj.ndim:
            if obj.ndim == 0:
                return obj[0]
            return np.asarray(obj)
        return np.ndarray.__array_wrap__(self, obj, context)

    def sliding_window(self, windowsize, stepsize=None):
        ws = tuple(self._compute_span(windowsize))
        ss = tuple(self._compute_span(stepsize)) if stepsize else ws
        result = sliding_window(self, ws, ss)

        try:
            new_dims = tuple(self._compute_new_dims(result, ws, ss))
        except ValueError:
            new_dims = [IdentityDimension()] * result.ndim

        return ArrayWithUnits(result, new_dims)

    def _sliding_window_integer_slices(self, windowsize, stepsize=None):
        windowsize = [windowsize] + [slice(None) for _ in self.dimensions[1:]]
        ws = tuple(self._compute_span(windowsize))
        if stepsize:
            stepsize = [stepsize] + [slice(None) for _ in self.dimensions[1:]]
            ss = tuple(self._compute_span(stepsize))
        else:
            ss = ws

        return ws, ss

    def sliding_window_with_leftovers(
            self, windowsize, stepsize=None, dopad=False):

        ws, ss = \
            self._sliding_window_integer_slices(windowsize, stepsize)

        leftovers, result = windowed(self, ws[0], ss[0], dopad)

        if not result.size:
            return self, ArrayWithUnits(result, self.dimensions)

        try:
            new_dims = tuple(self._compute_new_dims(result, ws, ss))
        except ValueError:
            new_dims = [IdentityDimension()] * result.ndim

        return leftovers, ArrayWithUnits(result, new_dims)

    def _compute_new_dims(self, windowed, ws, ss):
        for dimension, size, w, s in zip(self.dimensions, self.shape, ws, ss):
            try:
                modified = dimension.modified_dimension(size, w, s)
                for m in modified:
                    yield m
            except NotImplementedError:
                yield dimension

    def _compute_span(self, index):
        for sl in self._compute_indices(index):
            try:
                yield sl.stop - sl.start
            except (AttributeError, TypeError):
                yield sl

    def _compute_indices(self, index):
        dims_pos = 0
        for sl in index:
            if sl is None \
                    or isinstance(sl, int) \
                    or isinstance(sl, slice) \
                    or isinstance(sl, list):
                # burn one
                dims_pos += 1
                yield sl
            elif sl is Ellipsis:
                dims_pos += len(self.dimensions) - (len(index) - 1)
                yield Ellipsis
            else:
                dim = self.dimensions[dims_pos]
                yield dim.integer_based_slice(sl)
                dims_pos += 1

    def _new_dims(self, index, new_arr):
        dims_pos = 0
        shape_pos = 0
        not_ellipsis_or_none = len(filter(
            lambda x: x is not Ellipsis and x is not None, index))
        for sl in index:
            if sl is None:
                # additional dimension via np.newaxis
                yield IdentityDimension()
            elif isinstance(sl, int):
                # burn a dimension
                dims_pos += 1
            elif isinstance(sl, list):
                dims_pos += 1
                shape_pos += 1
                yield IdentityDimension()
            elif sl is Ellipsis:
                ellipsis_size = len(self.dimensions) - not_ellipsis_or_none
                for i in xrange(ellipsis_size):
                    yield self.dimensions[dims_pos]
                    dims_pos += 1
                    shape_pos += 1
            else:
                try:
                    dim = self.dimensions[dims_pos]
                    shape = new_arr.shape[shape_pos]
                    yield dim.metaslice(sl, shape)
                except IndexError:
                    yield dim
                dims_pos += 1
                shape_pos += 1

        # Return any leftover dimensions
        for dim in self.dimensions[dims_pos:]:
            yield dim

    def _wrap(self, other):
        return self.__class__(other, self.dimensions)

    def __gt__(self, other):
        return self._wrap(np.asarray(self).__gt__(other))

    def __ge__(self, other):
        return self._wrap(np.asarray(self).__ge__(other))

    def __lt__(self, other):
        return self._wrap(np.asarray(self).__lt__(other))

    def __le__(self, other):
        return self._wrap(np.asarray(self).__le__(other))

    def __eq__(self, other):
        return self._wrap(np.asarray(self).__eq__(other))

    def _tuplify(self, a):
        if isinstance(a, list):
            t = set(map(lambda x: x.__class__, a))
            if len(t) > 1:
                raise ValueError('a must be homogeneous')
            t = list(t)[0]
            if t == slice:
                return a
            else:
                return a,
        if isinstance(a, np.ndarray) and a.dtype == np.bool:
            return a,
        try:
            return tuple(a)
        except TypeError:
            return a,

    def __getslice__(self, i, j):
        return self.__getitem__(slice(i, j))

    def __setitem__(self, index, value):
        index = self._tuplify(index)
        indices = tuple(self._compute_indices(index))
        super(ArrayWithUnits, self).__setitem__(indices, value)

    def __getitem__(self, index):
        if isinstance(index, np.ndarray) \
                and index.dtype == np.bool:
            return np.asarray(self)[index]

        if self.ndim == 1 and isinstance(index, int):
            return np.asarray(self)[index]

        index = self._tuplify(index)
        indices = tuple(self._compute_indices(index))
        arr = super(ArrayWithUnits, self).__getitem__(indices)
        new_dims = tuple(self._new_dims(index, arr))
        return ArrayWithUnits(arr, new_dims)



from zounds.timeseries import TimeDimension
from basedimension import BaseDimensionEncoder, BaseDimensionDecoder
from util import encode_timedelta, decode_timedelta


class TimeDimensionEncoder(BaseDimensionEncoder):

    def __init__(self):
        super(TimeDimensionEncoder, self).__init__(TimeDimension)

    def dict(self, o):
        return dict(
                frequency=encode_timedelta(o.frequency),
                duration=encode_timedelta(o.duration),
                size=o.size)


class TimeDimensionDecoder(BaseDimensionDecoder):
    def __init__(self):
        super(TimeDimensionDecoder, self).__init__(TimeDimension)

    def kwargs(self, d):
        return dict(
            frequency=decode_timedelta(d['frequency']),
            duration=decode_timedelta(d['duration']),
            size=d['size'])



import unittest2
from random_samples import \
    ReservoirSampler, Reservoir, MultiplexedReservoir
from zounds.timeseries import TimeDimension, Seconds
from zounds.spectral import FrequencyDimension, FrequencyBand, LinearScale
from zounds.core import ArrayWithUnits, IdentityDimension
import numpy as np


class TestReservoir(unittest2.TestCase):
    def test_nsamples_must_be_gt_zero(self):
        self.assertRaises(ValueError, lambda: Reservoir(0))

    def test_can_dictate_dtype(self):
        r = Reservoir(100, dtype=np.float32)
        r.add(np.ones(10, dtype=np.float64))
        self.assertEqual(np.float32, r.get().dtype)

    def test_reservoir_has_first_input_dtype_when_unspecified(self):
        r = Reservoir(100)
        r.add(np.ones(10, dtype=np.float64))
        self.assertEqual(np.float64, r.get().dtype)

    def test_raises_if_nsamples_is_not_int(self):
        self.assertRaises(ValueError, lambda: Reservoir(1e2))

    def test_array_has_correct_first_dimension(self):
        r = Reservoir(100)
        r.add(np.random.random_sample((10, 3)))
        self.assertEqual(100, r.arr.shape[0])

    def test_can_add_samples_larger_than_reservoir_size(self):
        r = Reservoir(100)
        r.add(np.random.random_sample((1000, 3)))
        self.assertEqual(100, len(r.get()))

    def test_array_has_correct_subsequent_dimensions(self):
        r = Reservoir(100)
        r.add(np.random.random_sample((10, 3, 2)))
        self.assertEqual((3, 2), r.arr.shape[1:])

    def test_array_with_units(self):
        r = Reservoir(100)

        frequency_dimension = FrequencyDimension(
            LinearScale(FrequencyBand(100, 1000), 100))

        samples = ArrayWithUnits(
            np.ones((20, 100)),
            [
                TimeDimension(frequency=Seconds(1)),
                frequency_dimension
            ])

        r.add(samples)
        mixed = r.get()
        self.assertIsInstance(mixed, ArrayWithUnits)
        self.assertEqual(100, mixed.shape[1])
        self.assertIsInstance(mixed.dimensions[0], IdentityDimension)
        self.assertIsInstance(mixed.dimensions[1], FrequencyDimension)

    def test_reservoir_is_well_mixed(self):
        r = Reservoir(100)
        samples = np.arange(100)[..., None]
        for i in xrange(0, 100, 10):
            r.add(samples[i: i + 10])
        mixed = r.get().squeeze()
        diff = np.diff(mixed)
        self.assertFalse(np.all(diff == 1))

    def test_can_provide_explicit_indices_when_adding(self):
        r = Reservoir(10)
        samples = np.arange(10)[..., None]
        r.add(samples, indices=samples.squeeze()[::-1])
        mixed = r.get()
        np.testing.assert_allclose(mixed.squeeze(), samples.squeeze()[::-1])

    def test_raises_when_samples_and_explicit_indices_dont_match(self):
        r = Reservoir(10)
        samples = np.arange(10)[..., None]
        self.assertRaises(
            ValueError, lambda: r.add(samples, indices=samples.squeeze()[:5]))

    def test_can_get_batch(self):
        r = Reservoir(100)
        samples = np.arange(100)[..., None]
        for i in xrange(0, 100, 10):
            r.add(samples[i: i + 10])
        samples = r.get_batch(15)
        self.assertEqual(15, samples.shape[0])

    def test_raises_if_get_batch_is_larger_than_total_sample_size(self):
        r = Reservoir(100)
        samples = np.arange(100)[..., None]
        for i in xrange(0, 100, 10):
            r.add(samples[i: i + 10])
        self.assertRaises(ValueError, lambda: r.get_batch(1000))

    def test_raises_if_get_batch_is_larger_than_available_sample_size(self):
        r = Reservoir(100)
        samples = np.arange(100)[..., None]
        for i in xrange(0, 50, 10):
            r.add(samples[i: i + 10])
        self.assertRaises(ValueError, lambda: r.get_batch(64))


class TestMultiplexedReservoir(unittest2.TestCase):
    def test_is_consistent_across_keys(self):
        r = MultiplexedReservoir(100)
        samples = np.random.random_sample((10, 3))
        r.add(dict(cat=samples, dog=samples))
        mixed = r.get()
        np.testing.assert_allclose(mixed['cat'], mixed['dog'])

    def test_raises_when_wrong_set_of_keys_passed_to_add(self):
        r = MultiplexedReservoir(100)
        samples = np.random.random_sample((10, 3))
        r.add(dict(cat=samples, dog=samples))
        self.assertRaises(
            ValueError, lambda: r.add(dict(rat=samples, frog=samples)))

    def test_raises_when_some_keys_have_mismatched_lengths(self):
        r = MultiplexedReservoir(100)
        samples = np.random.random_sample((10, 3))
        self.assertRaises(
            ValueError, lambda: r.add(dict(cat=samples, dog=samples[:-1])))

    def test_raises_when_some_keys_have_mismatched_lengths_second_add(self):
        r = MultiplexedReservoir(100)
        samples = np.random.random_sample((10, 3))
        r.add(dict(cat=samples, dog=samples))
        self.assertRaises(
            ValueError, lambda: r.add(dict(cat=samples, dog=samples[:-1])))

    def test_get_returns_dict_with_user_specified_keys(self):
        r = MultiplexedReservoir(100)
        samples = np.random.random_sample((10, 3))
        d = dict(cat=samples, dog=samples)
        r.add(d)
        mixed = r.get()
        self.assertEqual(set(d.keys()), set(mixed.keys()))


class TestReservoirSampler(unittest2.TestCase):
    def test_can_sample_from_one_dimensional_feature(self):
        sampler = ReservoirSampler(nsamples=10)

        frequency_dimension = FrequencyDimension(
            LinearScale(FrequencyBand(100, 1000), 100))

        samples = ArrayWithUnits(
            np.ones((20, 100)),
            [
                TimeDimension(frequency=Seconds(1)),
                frequency_dimension
            ])

        sampler._enqueue(samples, pusher=None)
        reservoir = sampler._r
        self.assertEqual((10, 100), reservoir.shape)
        self.assertIsInstance(reservoir, ArrayWithUnits)
        self.assertEqual(reservoir.dimensions[0], IdentityDimension())
        self.assertEqual(reservoir.dimensions[1], frequency_dimension)

    def test_can_wrap_samples(self):
        sampler = ReservoirSampler(nsamples=10)

        frequency_dimension = FrequencyDimension(
            LinearScale(FrequencyBand(100, 1000), 100))

        samples = ArrayWithUnits(
            np.ones((2, 10, 100)),
            [
                TimeDimension(frequency=Seconds(10)),
                TimeDimension(frequency=Seconds(1)),
                frequency_dimension
            ])

        sampler._enqueue(samples, pusher=None)
        reservoir = sampler._r
        self.assertEqual((10, 10, 100), reservoir.shape)
        self.assertIsInstance(reservoir, ArrayWithUnits)
        self.assertEqual(reservoir.dimensions[0], IdentityDimension())
        self.assertEqual(reservoir.dimensions[1], samples.dimensions[1])
        self.assertEqual(reservoir.dimensions[2], samples.dimensions[2])

    def test_can_dequeue_when_reservoir_is_full(self):
        sampler = ReservoirSampler(nsamples=10)

        frequency_dimension = FrequencyDimension(
            LinearScale(FrequencyBand(100, 1000), 100))

        samples = ArrayWithUnits(
            np.ones((10, 10, 100)),
            [
                TimeDimension(frequency=Seconds(10)),
                TimeDimension(frequency=Seconds(1)),
                frequency_dimension
            ])

        sampler._enqueue(samples, pusher=None)
        reservoir = sampler._dequeue()

        self.assertEqual((10, 10, 100), reservoir.shape)
        self.assertIsInstance(reservoir, ArrayWithUnits)
        self.assertEqual(reservoir.dimensions[0], IdentityDimension())
        self.assertEqual(reservoir.dimensions[1], samples.dimensions[1])
        self.assertEqual(reservoir.dimensions[2], samples.dimensions[2])

    def test_can_dequeue_when_reservoir_is_partially_full(self):
        sampler = ReservoirSampler(nsamples=10)

        frequency_dimension = FrequencyDimension(
            LinearScale(FrequencyBand(100, 1000), 100))

        samples = ArrayWithUnits(
            np.ones((4, 10, 100)),
            [
                TimeDimension(frequency=Seconds(10)),
                TimeDimension(frequency=Seconds(1)),
                frequency_dimension
            ])

        sampler._enqueue(samples, pusher=None)
        reservoir = sampler._dequeue()

        self.assertEqual((4, 10, 100), reservoir.shape)
        self.assertIsInstance(reservoir, ArrayWithUnits)
        self.assertEqual(reservoir.dimensions[0], IdentityDimension())
        self.assertEqual(reservoir.dimensions[1], samples.dimensions[1])
        self.assertEqual(reservoir.dimensions[2], samples.dimensions[2])



import json
import os
import threading
import numpy as np
from hammingdb import HammingDb
from zounds.persistence import TimeSliceEncoder, TimeSliceDecoder
from zounds.timeseries import ConstantRateTimeSeries
from zounds.timeseries import TimeSlice


class SearchResults(object):
    def __init__(self, query, results):
        super(SearchResults, self).__init__()
        self.query = query
        self.results = results

    def __iter__(self):
        for result in self.results:
            yield result


class HammingIndex(object):
    def __init__(
            self,
            document,
            feature,
            version=None,
            path='',
            db_size_bytes=1000000000,
            listen=False):

        super(HammingIndex, self).__init__()
        self.document = document
        self.feature = feature
        self.db_size_bytes = db_size_bytes
        self.path = path

        version = version or self.feature.version

        self.hamming_db_path = os.path.join(
            self.path, 'index.{self.feature.key}.{version}'
                .format(**locals()))

        try:
            self.event_log = document.event_log
        except AttributeError:
            self.event_log = None

        try:
            self.hamming_db = HammingDb(self.hamming_db_path, code_size=None)
        except ValueError:
            self.hamming_db = None

        self.encoder = TimeSliceEncoder()
        self.decoder = TimeSliceDecoder()
        self.thread = None

        if listen:
            self.listen()

    def close(self):
        try:
            self.stop()
        except:
            pass

        try:
            self.hamming_db.close()
        except:
            pass

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __len__(self):
        return len(self.hamming_db)

    def stop(self):
        self.event_log.unsubscribe()

    def listen(self):
        self.thread = threading.Thread(target=self._listen)
        self.thread.daemon = True
        self.thread.start()

    def _init_hamming_db(self, code):
        if self.hamming_db is not None:
            return
        self.hamming_db = HammingDb(self.hamming_db_path, code_size=len(code))

    def _synchronously_process_events(self):
        self._listen(raise_when_empty=True)

    def add_all(self):
        for doc in self.document:
            self.add(doc._id)

    def add(self, _id, timestamp=''):
        # load the feature from the feature database
        feature = self.feature(_id=_id, persistence=self.document)

        try:
            arr = ConstantRateTimeSeries(feature)
        except ValueError:
            arr = feature

        # extract codes and timeslices from the feature
        for ts, data in arr.iter_slices():
            code = self.encode_query(data)
            encoded_ts = dict(
                _id=_id,
                **self.encoder.dict(ts))
            self._init_hamming_db(code)
            self.hamming_db.append(code, json.dumps(encoded_ts))
            self.hamming_db.set_metadata('timestamp', bytes(timestamp))

    def _listen(self, raise_when_empty=False):

        if self.hamming_db is not None:
            last_timestamp = self.hamming_db.get_metadata('timestamp') or ''
        else:
            last_timestamp = ''

        if not self.event_log:
            raise ValueError(
                '{self.document} must have an event log configured'
                    .format(**locals()))

        subscription = self.event_log.subscribe(
            last_id=last_timestamp, raise_when_empty=raise_when_empty)

        for timestamp, data in subscription:

            # parse the data from the event stream
            data = json.loads(data)
            _id, name, version = data['_id'], data['name'], data['version']

            # ensure that it's about the feature we're subscribed to
            if name != self.feature.key or version != self.feature.version:
                continue

            self.add(_id, timestamp)

    def _parse_result(self, result):
        d = json.loads(result)
        ts = TimeSlice(**self.decoder.kwargs(d))
        return d['_id'], ts

    def decode_query(self, binary_query):
        packed = np.fromstring(binary_query, dtype=np.uint8)
        return np.unpackbits(packed)

    def encode_query(self, feature):
        try:
            return np.packbits(feature).tostring()
        except TypeError:
            # the query is already packed and expressed as raw bytes
            return feature

    def random_search(self, n_results, multithreaded=False, sort=False):
        code, raw_results = self.hamming_db.random_search(
            n_results, multithreaded, sort=sort)
        parsed_results = (self._parse_result(r) for r in raw_results)
        return SearchResults(code, parsed_results)

    def search(self, feature, n_results, multithreaded=False, sort=False):
        code = self.encode_query(feature)
        raw_results = self.hamming_db.search(
            code, n_results, multithreaded, sort=sort)
        parsed_results = (self._parse_result(r) for r in raw_results)
        return SearchResults(code, parsed_results)



import lmdb
from zounds.nputil import Growable, packed_hamming_distance
import numpy as np
from multiprocessing.dummy import Pool as ThreadPool
from multiprocessing import cpu_count
import os
import binascii


class HammingDb(object):
    def __init__(self, path, map_size=1000000000, code_size=8):
        super(HammingDb, self).__init__()

        if not os.path.exists(path):
            os.makedirs(path)

        self.path = path
        self.env = lmdb.open(
            self.path,
            max_dbs=10,
            map_size=map_size,
            writemap=True,
            map_async=True,
            metasync=True)

        self.metadata = self.env.open_db('metadata')
        try:
            self.code_size = int(self.get_metadata('codesize'))
            if code_size and code_size != self.code_size:
                raise ValueError(
                    'Database is already initialized with code size {code_size}'
                    ', but {self.code_size} was passed to __init__'
                        .format(**locals()))
        except TypeError:
            if code_size is None:
                raise ValueError(
                    'You must supply a code size for an uninitialized database')
            if code_size % 8:
                raise ValueError('code_size must be a multiple of 8')
            self.set_metadata('codesize', str(code_size))
            self.code_size = code_size

        self.index = self.env.open_db('index')
        self._append_buffer = self._recarray(1)
        self._code_bytearray = bytearray('a' * self.code_size)
        self._code_buffer = np.frombuffer(self._code_bytearray, dtype=np.uint64)
        self._codes = None
        self._ids = set()
        self._catch_up_on_in_memory_store()

        self._thread_count = cpu_count()
        self._pool = ThreadPool(processes=self._thread_count)

    def close(self):
        self.env.close()

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def set_metadata(self, key, value):
        with self.env.begin(write=True) as txn:
            txn.put(key, value, db=self.metadata)

    def get_metadata(self, key):
        with self.env.begin() as txn:
            return txn.get(key, db=self.metadata)

    def _catch_up_on_in_memory_store(self):
        self._initialize_in_memory_store()
        with self.env.begin() as txn:
            cursor = txn.cursor(db=self.index)
            for i, bundle in enumerate(cursor.iternext(keys=True, values=True)):
                _id, value = bundle
                if _id in self._ids:
                    continue

                code = value[:self.code_size]
                self._add_code(_id, code)

    def __len__(self):
        with self.env.begin() as txn:
            lmdb_size = txn.stat(self.index)['entries']
            if not lmdb_size:
                return 0
            return lmdb_size

    def _recarray(self, size):
        return np.recarray(
            size,
            dtype=[
                ('id', 'S32'),
                ('code', np.uint64, self.code_size // 8)],
            order='F')

    def _initialize_in_memory_store(self):
        if self._codes is not None:
            return
        self._codes = Growable(self._recarray(int(1e6)))

    def _np_code(self, code):
        self._code_bytearray[:] = code
        return self._code_buffer

    def _validate_code_size(self, code):
        code_len = len(code)
        if code_len != self.code_size:
            fmt = '''code must be equal to code_size
                    ({self.code_size}), but was {code_len}'''
            raise ValueError(fmt.format(**locals()))

    def _add_code(self, _id, code):
        arr = self._append_buffer
        arr[0]['id'] = _id
        arr[0]['code'] = self._np_code(code)
        self._codes.append(arr)
        self._ids.add(_id)

    def _check_for_external_modifications(self):
        if self.__len__() != self._codes.logical_size:
            self._catch_up_on_in_memory_store()

    def _new_id(self):
        return binascii.hexlify(os.urandom(16))

    def append(self, code, data):
        self._validate_code_size(code)
        self._initialize_in_memory_store()

        with self.env.begin(write=True) as txn:
            _id = self._new_id()
            txn.put(_id, code + data, db=self.index)
            self._add_code(_id, code)

    def _random_code(self):
        with self.env.begin() as txn:
            with txn.cursor(self.index) as cursor:
                code = None
                while not code:
                    if cursor.set_range(self._new_id()):
                        return txn.get(
                            cursor.key(), db=self.index)[:self.code_size]
                    continue

    def random_search(self, n_results, multithreaded=False, sort=False):
        code = self._random_code()
        return code, self.search(code, n_results, multithreaded, sort=sort)

    def search(self, code, n_results, multithreaded=False, sort=False):
        self._validate_code_size(code)
        self._check_for_external_modifications()
        query = self._np_code(code)

        codes = self._codes.logical_data['code']

        if codes.ndim == 1:
            codes = codes[..., None]

        if not multithreaded:
            scores = packed_hamming_distance(query, codes)
        else:
            n_codes = len(codes)
            chunksize = max(1, n_codes // self._thread_count)
            scores = np.concatenate(self._pool.map(
                lambda x: packed_hamming_distance(query, x),
                (codes[i: i + chunksize] for i in
                 xrange(0, n_codes, chunksize))))

        # indices = np.argsort(scores)[:n_results]

        # argpartition will ensure that the lowest scores will all be
        # withing the first n_results elements, but makes no guarantees
        # about the ordering *within* n_results
        partitioned_indices = np.argpartition(scores, n_results)[:n_results]

        if sort:
            # since argpartition doesn't guarantee that the results are
            # sorted *within* n_results, sort the much smaller result set
            sorted_indices = np.argsort(scores[partitioned_indices])

            indices = partitioned_indices[sorted_indices]
        else:
            # the partitioned indices are good enough.  results will all be
            # within some degree of similarity, but not necessarily in any
            # particular order
            indices = partitioned_indices

        nearest = self._codes.logical_data[indices]['id']

        with self.env.begin() as txn:
            for _id in nearest:
                yield txn.get(_id, db=self.index)[self.code_size:]



import unittest2
from timedimension import TimeDimensionEncoder, TimeDimensionDecoder
from zounds.timeseries import TimeDimension, Seconds, Milliseconds


class TimeDimensionTests(unittest2.TestCase):
    def setUp(self):
        self.encoder = TimeDimensionEncoder()
        self.decoder = TimeDimensionDecoder()

    def test_can_round_trip(self):
        td = TimeDimension(Seconds(1), Milliseconds(500), 100)
        encoded = self.encoder.encode(td)
        decoded = self.decoder.decode(encoded)
        self.assertIsInstance(decoded, TimeDimension)
        self.assertEqual(Seconds(1), decoded.frequency)
        self.assertEqual(Milliseconds(500), decoded.duration)
        self.assertEqual(100, decoded.size)



"""
The core module introduces the key building blocks of the representations zounds
deals in: :class:`ArrayWithUnits`, a :class:`numpy.ndarray`-derived class that
supports semantically meaningful indexing, and :class:`Dimension`, a common
base class for custom, user-defined dimensions.
"""

from dimensions import Dimension, IdentityDimension
from axis import ArrayWithUnits

__all__ = [Dimension, IdentityDimension, ArrayWithUnits]



"""
The spectral module contains classes that aid in dealing with frequency-domain
representations of sound
"""

from sliding_window import \
    SlidingWindow, OggVorbisWindowingFunc, WindowingFunc, HanningWindowingFunc

from spectral import \
    FFT, DCT, DCTIV, MDCT, BarkBands, Chroma, BFCC, SpectralCentroid, \
    SpectralFlatness, FrequencyAdaptiveTransform, FrequencyWeighting

from tfrepresentation import FrequencyDimension, ExplicitFrequencyDimension

from weighting import AWeighting

from frequencyscale import \
    LinearScale, LogScale, FrequencyBand, FrequencyScale, GeometricScale, \
    ExplicitScale

from frequencyadaptive import FrequencyAdaptive

from functional import fft, stft, apply_scale, frequency_decomposition


from basedimension import BaseDimensionEncoder, BaseDimensionDecoder
from zounds.core import IdentityDimension


class IdentityDimensionEncoder(BaseDimensionEncoder):
    def __init__(self):
        super(IdentityDimensionEncoder, self).__init__(IdentityDimension)

    def dict(self, o):
        return dict()


class IdentityDimensionDecoder(BaseDimensionDecoder):
    def __init__(self):
        super(IdentityDimensionDecoder, self).__init__(IdentityDimension)



import numpy as np
import unittest2

from synthesize import \
    SineSynthesizer, DCTSynthesizer, FFTSynthesizer, NoiseSynthesizer, \
    SilenceSynthesizer, FrequencyDecompositionSynthesizer
from zounds.basic import stft, resampled
from zounds.core import ArrayWithUnits
from zounds.persistence import ArrayWithUnitsFeature
from zounds.spectral import \
    FrequencyDimension, FrequencyBand, LinearScale, FFT, SlidingWindow, \
    OggVorbisWindowingFunc, frequency_decomposition
from zounds.timeseries import \
    SR22050, SR44100, SR11025, SR48000, SR96000, HalfLapped, Seconds, \
    TimeDimension, AudioSamples, SampleRate, Milliseconds, TimeSlice
from zounds.util import simple_in_memory_settings


class SynthesizeTests(unittest2.TestCase):
    def test_has_correct_sample_rate(self):
        half_lapped = HalfLapped()
        synth = DCTSynthesizer()
        raw = np.zeros((100, 2048))
        band = FrequencyBand(0, SR44100().nyquist)
        scale = LinearScale(band, raw.shape[1])
        timeseries = ArrayWithUnits(
            raw, [TimeDimension(*half_lapped), FrequencyDimension(scale)])
        output = synth.synthesize(timeseries)
        self.assertIsInstance(output.samplerate, SR44100)
        self.assertIsInstance(output, AudioSamples)


class FFTSynthesizerTests(unittest2.TestCase):
    def can_invert_fft(self, samplerate):
        base_cls = stft(
            resample_to=samplerate,
            store_fft=True,
            store_windowed=True)

        @simple_in_memory_settings
        class Document(base_cls):
            pass

        synth = SineSynthesizer(samplerate)
        audio = synth.synthesize(Seconds(2), freqs_in_hz=[440., 880.])

        _id = Document.process(meta=audio.encode())
        doc = Document(_id)

        fft_synth = FFTSynthesizer()
        recon = fft_synth.synthesize(doc.fft)

        self.assertIsInstance(recon, ArrayWithUnits)
        self.assertEqual(audio.dimensions, recon.dimensions)

    def test_can_invert_long_fft(self):
        samplerate = SR11025()
        rs = resampled(resample_to=samplerate)

        @simple_in_memory_settings
        class Document(rs):
            long_windowed = ArrayWithUnitsFeature(
                SlidingWindow,
                wscheme=SampleRate(
                    Milliseconds(500),
                    Seconds(1)),
                wfunc=OggVorbisWindowingFunc(),
                needs=rs.resampled,
                store=True)

            long_fft = ArrayWithUnitsFeature(
                FFT,
                needs=long_windowed,
                store=True)

        synth = SineSynthesizer(samplerate)
        audio = synth.synthesize(Seconds(2), freqs_in_hz=[440., 880.])

        _id = Document.process(meta=audio.encode())
        doc = Document(_id)

        fft_synth = FFTSynthesizer()
        recon = fft_synth.synthesize(doc.long_fft)
        self.assertIsInstance(recon, AudioSamples)
        self.assertEqual(audio.dimensions, recon.dimensions)

    def test_can_invert_fft_11025(self):
        self.can_invert_fft(SR11025())

    def test_can_invert_fft_22050(self):
        self.can_invert_fft(SR22050())

    def test_can_invert_fft_44100(self):
        self.can_invert_fft(SR44100())

    def test_can_invert_fft_48000(self):
        self.can_invert_fft(SR48000())

    @unittest2.skip(
        'HalfLapped does not compute the right window size in samples')
    def test_can_invert_fft_96000(self):
        self.can_invert_fft(SR96000())


class SineSynthesizerTests(unittest2.TestCase):
    def test_generates_correct_shape(self):
        ss = SineSynthesizer(SR22050())
        audio = ss.synthesize(Seconds(4), freqs_in_hz=[440.])
        self.assertEqual(1, len(audio.shape))

    def test_generates_correct_samplerate(self):
        ss = SineSynthesizer(SR44100())
        audio = ss.synthesize(Seconds(4), freqs_in_hz=[440.])
        self.assertEqual(SR44100(), audio.samplerate)

    def test_generates_correct_number_of_samples(self):
        samplerate = SR22050()
        duration = Seconds(1)
        ss = SineSynthesizer(samplerate)
        audio = ss.synthesize(duration, freqs_in_hz=[440.])
        expected_samples = int(samplerate) * int(duration / Seconds(1))
        self.assertEqual(expected_samples, len(audio))

    def test_can_create_audio_with_single_tone(self):
        ss = SineSynthesizer(SR22050())
        audio = ss.synthesize(Seconds(4), freqs_in_hz=[440.])
        fft = abs(np.fft.rfft(audio))
        self.assertEqual(1, (fft > 1).sum())

    def test_can_create_audio_with_multiple_tones(self):
        ss = SineSynthesizer(SR22050())
        audio = ss.synthesize(Seconds(4), freqs_in_hz=[440., 660.])
        fft = abs(np.fft.rfft(audio))
        self.assertEqual(2, (fft > 1).sum())


class NoiseSynthesizerTests(unittest2.TestCase):
    def test_noise_synth_outputs_values_in_correct_range(self):
        ns = NoiseSynthesizer(SR11025())
        audio = ns.synthesize(Seconds(1))
        self.assertLess(audio.min(), 0)
        self.assertGreater(audio.max(), 0)


class SilenceSynthesizerTests(unittest2.TestCase):
    def test_silence_synthesizer_outputs_zero(self):
        synth = SilenceSynthesizer(SR11025())
        audio = synth.synthesize(Seconds(1))
        np.testing.assert_allclose(audio, 0)


class FrequencyDecompositionSynthesizerTests(unittest2.TestCase):
    def test_can_resynthesize_frequency_decomposition(self):
        sr = SR22050()
        samples = SilenceSynthesizer(sr).synthesize(Milliseconds(9999))
        window_size = 8192

        wscheme = sr.windowing_scheme(window_size, window_size // 2)
        duration = TimeSlice(wscheme.duration)
        frequency = TimeSlice(wscheme.frequency)
        _, windowed = samples.sliding_window_with_leftovers(
            duration, frequency, dopad=True)
        fa = frequency_decomposition(
            windowed, [32, 64, 128, 256, 512, 1024, 2048, 4096])

        fdsynth = FrequencyDecompositionSynthesizer(sr, window_size)

        samples = fdsynth.synthesize(fa)
        self.assertEqual(2, samples.ndim)
        self.assertEqual(windowed.dimensions[1], samples.dimensions[1])
        self.assertEqual(windowed.dimensions[0], samples.dimensions[0])



import requests
import urlparse
from zounds.soundfile.audio_metadata import AudioMetaData


class InternetArchive(object):
    """
    Produces an iterable of :class:`zounds.soundfile.AudioMetaData` instances
    for every file of a particular format from an internet archive id.

    Args:
        archive_id (str): the Internet Archive identifier
        format_filter (str): The file format to return
        attrs (dict): Extra attributes to add to the :class:`AudioMetaData`

    Raises:
        ValueError: when archive_id is not provided

    Examples:
        >>> from zounds import InternetArchive
        >>> ia = InternetArchive('Greatest_Speeches_of_the_20th_Century')
        >>> iter(ia).next()
        {'creator': u'John F. Kennedy', 'height': u'0', 'channels': None, 'genre': u'Folk', 'licensing': None, 'mtime': u'1236666800', 'samplerate': None, 'size': u'7264435', 'album': u'Great Speeches of the 20th Century [Box Set] Disc 2', 'title': u'The Cuban Missile Crisis', 'format': u'128Kbps MP3', 'source': u'original', 'description': None, 'tags': None, 'track': u'15', 'crc32': u'ace17eb5', 'md5': u'e00f4e7bd9df7bdba4db7098d1ccdfe0', 'sha1': u'e42d1f348078a11ed9a6ea9c8934a1236235c7b3', 'artist': u'John F. Kennedy', 'external-identifier': [u'urn:acoustid:ff850a0c-2efa-450f-8034-efdb31a9b696', u'urn:mb_recording_id:912cedd0-5530-4f26-972c-13d131fef06e'], 'uri': <Request [GET]>, 'length': u'454.03', 'width': u'0'}

    See Also:
        :class:`FreeSoundSearch`
        :class:`PhatDrumLoops`
        :class:`zounds.soundfile.AudioMetaData`
    """
    def __init__(self, archive_id, format_filter=None, **attrs):
        super(InternetArchive, self).__init__()

        self.attrs = attrs
        if not archive_id:
            raise ValueError('You must supply an Internet Archive id')

        self.format_filter = format_filter or \
            (lambda x: x['format'] == 'Ogg Vorbis')
        self.archive_id = archive_id

    def _get_metadata(self, data, all_files):
        if data['source'] == 'original':
            return data
        elif data['source'] == 'derivative':
            return all_files['/' + data['original']]

    def __iter__(self):
        base_url = 'https://archive.org/'
        archive_id = self.archive_id
        url = urlparse.urljoin(
            base_url, '/details/{archive_id}&output=json'.format(**locals()))
        resp = requests.get(url)

        all_files = resp.json()['files']
        for k, v in all_files.iteritems():
            if self.format_filter(v):
                sound_url = urlparse.urljoin(
                    base_url, '/download/{archive_id}{k}'.format(**locals()))
                request = requests.Request(method='GET', url=sound_url)
                metadata = self._get_metadata(v, all_files)
                metadata.update(self.attrs)
                web_url = 'https://archive.org//details/{archive_id}'\
                    .format(**locals())
                metadata.update(web_url=web_url)
                yield AudioMetaData(uri=request, **metadata)



import unittest2
from zounds.util import simple_in_memory_settings
import featureflow as ff
from preprocess import Log, PreprocessingPipeline
import numpy as np


class LogTests(unittest2.TestCase):
    def get_model(self):
        @simple_in_memory_settings
        class Model(ff.BaseModel):
            log = ff.PickleFeature(
                    Log,
                    store=False)

            pipeline = ff.PickleFeature(
                    PreprocessingPipeline,
                    needs=(log,),
                    store=True)

        training = np.random.random_sample((10, 3))
        _id = Model.process(log=training)
        return Model(_id)

    def test_forward_transform_preserves_sign_large_values(self):
        model = self.get_model()
        inp = (np.random.random_sample((100, 3)) * 1000) - 500
        s1 = np.sign(inp)
        result = model.pipeline.transform(inp)
        s2 = np.sign(result.data)
        self.assertTrue(np.all(s1 == s2))

    def test_forward_transform_preserves_sign(self):
        model = self.get_model()
        inp = np.random.random_sample((100, 3)) - 0.5
        s1 = np.sign(inp)
        result = model.pipeline.transform(inp)
        s2 = np.sign(result.data)
        self.assertTrue(np.all(s1 == s2))

    def test_backward_transform_reconstructs_data_large_values(self):
        model = self.get_model()
        inp = (np.random.random_sample((100, 3)) * 1000) - 500
        result = model.pipeline.transform(inp)
        inverted = result.inverse_transform()
        np.testing.assert_allclose(inp, inverted)

    def test_backward_transform_reconstructs_data(self):
        model = self.get_model()
        inp = np.random.random_sample((100, 3)) - 0.5
        result = model.pipeline.transform(inp)
        inverted = result.inverse_transform()
        np.testing.assert_allclose(inp, inverted)

    def test_pipeline_requires_no_inversion_data(self):
        model = self.get_model()
        inp = (np.random.random_sample((100, 3)) * 1000) - 500
        result = model.pipeline.transform(inp)
        self.assertFalse(result.inversion_data[0])



"""
The datasets module provides access to some common sources of audio on the
internet.  In general, a dataset instance is an iterable of
:class:`zounds.soundfile.AudioMetaData` instances that can be passed to the
root node of an audio processing graph.
"""

from musicnet import MusicNet
from phatdrumloops import PhatDrumLoops
from internetarchive import InternetArchive
from freesound import FreeSoundSearch
from filesystem import Directory
from cache import DataSetCache
from ingest import ingest
from predownload import PreDownload
from nsynth import NSynth



import unittest2
from timeslice import TimeSliceEncoder, TimeSliceDecoder
from zounds.timeseries import TimeSlice, Seconds, Milliseconds


class TimeSliceRoundTripTests(unittest2.TestCase):
    def _roundtrip(self, ts):
        encoder = TimeSliceEncoder()
        decoder = TimeSliceDecoder()
        encoded = encoder.dict(ts)
        decoded = TimeSlice(**decoder.kwargs(encoded))
        self.assertEqual(ts, decoded)

    def test_can_roundtrip(self):
        self._roundtrip(TimeSlice(start=Seconds(1), duration=Seconds(1)))

    def test_can_roundtrip_milliseconds(self):
        self._roundtrip(TimeSlice(start=Seconds(1), duration=Milliseconds(250)))



from scipy.cluster.vq import kmeans
from featureflow import Node
from preprocess import PreprocessResult, Preprocessor


class KMeans(Preprocessor):
    """

    """
    def __init__(self, centroids=None, needs=None):
        super(KMeans, self).__init__(needs=needs)
        self._centroids = centroids

    def _forward_func(self):
        def x(d, codebook=None):
            import numpy as np
            from scipy.spatial.distance import cdist
            from zounds.core import ArrayWithUnits, IdentityDimension
            l = d.shape[0]
            dist = cdist(d, codebook)
            best = np.argmin(dist, axis=1)
            feature = np.zeros((l, len(codebook)), dtype=np.uint8)
            feature[np.arange(l), best] = 1
            try:
                return ArrayWithUnits(
                    feature, [d.dimensions[0], IdentityDimension()])
            except AttributeError:
                return feature

        return x

    def _backward_func(self):
        def x(d, codebook=None):
            import numpy as np
            indices = np.where(d == 1)
            return codebook[indices[1]]

        return x

    def _process(self, data):
        data = self._extract_data(data)
        codebook, _ = kmeans(data, self._centroids)
        op = self.transform(codebook=codebook)
        inv_data = self.inversion_data(codebook=codebook)
        inv = self.inverse_transform()
        data = op(data)
        yield PreprocessResult(
            data, op, inversion_data=inv_data, inverse=inv, name='KMeans')


class Learned(Node):
    """

    """
    def __init__(
            self, 
            learned=None, 
            version=None, 
            wrapper=None,
            pipeline_func=None,
            needs=None):
        super(Learned, self).__init__(needs=needs)
        self._pipeline_func = pipeline_func or (lambda x: x.pipeline)
        self._wrapper = wrapper
        self._learned = learned
        self._version = version

    @property
    def version(self):
        if self._version is not None:
            return self._version

        pipeline = self._pipeline_func(self._learned)
        return pipeline.version

    def _process(self, data):
        pipeline = self._pipeline_func(self._learned)
        transformed = pipeline.transform(data, wrapper=self._wrapper).data
        yield transformed



import unittest2
import featureflow as ff
import numpy as np

from zounds.util import simple_in_memory_settings
from preprocess import Reshape, PreprocessingPipeline


class ReshapeTests(unittest2.TestCase):

    def do_setup(self, shape, new_shape):
        @simple_in_memory_settings
        class Model(ff.BaseModel):
            flattened = ff.PickleFeature(
                    Reshape,
                    new_shape=new_shape,
                    store=False)

            pipeline = ff.PickleFeature(
                    PreprocessingPipeline,
                    needs=(flattened,),
                    store=True)

        training = np.random.random_sample(shape)
        _id = Model.process(flattened=training)
        model = Model(_id)
        data = np.random.random_sample(shape)
        transformed = model.pipeline.transform(data)
        inverted = transformed.inverse_transform()
        return data, transformed.data, inverted

    def do_assertions(self, original_shape, new_shape, expected_shape):
        data, transformed, inverted = self.do_setup(original_shape, new_shape)
        self.assertEqual(expected_shape, transformed.shape)
        self.assertEqual(original_shape, inverted.shape)
        np.testing.assert_allclose(data.ravel(), transformed.ravel())
        np.testing.assert_allclose(data.ravel(), inverted.ravel())

    def test_reshape(self):
        self.do_assertions(
            original_shape=(10, 100),
            new_shape=(10, 10),
            expected_shape=(10, 10, 10))

    def test_expand_multiple_dimensions(self):
        self.do_assertions(
            original_shape=(10, 40),
            new_shape=(1, -1, 1),
            expected_shape=(10, 1, 40, 1))

    def test_expand_middle_dimension(self):
        self.do_assertions(
            original_shape=(10, 40),
            new_shape=(1, -1),
            expected_shape=(10, 1, 40))

    def test_expand_last_dimension(self):
        self.do_assertions(
            original_shape=(10, 40),
            new_shape=(-1, 1),
            expected_shape=(10, 40, 1))

    def test_flatten_one_dimensional(self):
        self.do_assertions(
            original_shape=(10,),
            new_shape=tuple(),
            expected_shape=(10,))

    def test_flatten_two_dimensional(self):
        self.do_assertions(
            original_shape=(10, 40),
            new_shape=(-1,),
            expected_shape=(10, 40))

    def test_flatten_three_dimensional(self):
        self.do_assertions(
            original_shape=(10, 40, 20),
            new_shape=(-1,),
            expected_shape=(10, 800))



import numpy as np
import featureflow as ff
from random_samples import ShuffledSamples


def learning_pipeline(dtype=np.float32):
    class LearningPipeline(ff.BaseModel):
        samples = ff.PickleFeature(ff.IteratorNode)

        shuffled = ff.PickleFeature(
            ShuffledSamples,
            nsamples=ff.Var('nsamples'),
            dtype=dtype,
            needs=samples)

    return LearningPipeline



import unittest2
import numpy as np
from dimensions import Dimension, IdentityDimension
from axis import ArrayWithUnits, CustomSlice
from string import ascii_lowercase


class ContrivedSlice(CustomSlice):
    def __init__(self, start=None, stop=None, step=None):
        super(ContrivedSlice, self).__init__()
        self.step = step
        self.stop = stop
        self.start = start


class ContrivedDimension(Dimension):
    def __init__(self, factor):
        super(ContrivedDimension, self).__init__()
        self.factor = factor
        self.size = None

    def copy(self):
        return ContrivedDimension(self.factor)

    def modified_dimension(self, size, windowsize, stepsize=None):
        yield ContrivedDimension(self.factor * windowsize)
        yield ContrivedDimension(self.factor)

    def integer_based_slice(self, index):
        if isinstance(index, ContrivedSlice):
            return slice(index.start // self.factor, index.stop // self.factor)
        else:
            return index

    def __eq__(self, other):
        try:
            return self.factor == other.factor
        except AttributeError:
            return False

    def __repr__(self):
        return 'ContrivedDimension(factor={factor})'.format(**self.__dict__)


class ContrivedDimension2(Dimension):
    def __init__(self, factor):
        super(ContrivedDimension2, self).__init__()
        self.factor = factor
        self.size = None

    def copy(self):
        return ContrivedDimension2(self.factor)

    def modified_dimension(self, size, windowsize, stepsize=None):
        yield ContrivedDimension(self.factor * windowsize)
        yield ContrivedDimension(self.factor)

    def integer_based_slice(self, index):
        if isinstance(index, ContrivedSlice):
            return slice(index.start // self.factor, index.stop // self.factor)
        else:
            return index

    def __eq__(self, other):
        try:
            return self.factor == other.factor
        except AttributeError:
            return False

    def __repr__(self):
        return 'ContrivedDimension2(factor={factor})'.format(**self.__dict__)


class AsciiCharacterDimension(Dimension):
    def __init__(self, labels):
        super(AsciiCharacterDimension, self).__init__()
        self.labels = labels

    def copy(self):
        return AsciiCharacterDimension(self.labels)

    def modified_dimension(self, size, windowsize, stepsize=None):
        raise NotImplementedError()

    def metaslice(self, index, size):
        return AsciiCharacterDimension(self.labels[index])

    def integer_based_slice(self, index):
        return index


class ContrivedArray(ArrayWithUnits):
    def __new__(cls, arr, dimensions):
        return ArrayWithUnits.__new__(cls, arr, dimensions)


class CoreTests(unittest2.TestCase):
    def test_maintain_array_with_units_with_boolean_condition(self):
        arr = ArrayWithUnits(
            np.random.random_sample((10, 100)) - 0.5,
            [ContrivedDimension(10), ContrivedDimension2(10)])
        binary = arr >= 0
        self.assertIsInstance(binary, ArrayWithUnits)
        self.assertIsInstance(arr.dimensions[0], ContrivedDimension)
        self.assertIsInstance(arr.dimensions[1], ContrivedDimension2)

    def test_can_reshape_and_downgrade_to_numpy_array(self):
        arr = ArrayWithUnits(
            np.zeros((100, 10)),
            [ContrivedDimension(10), ContrivedDimension2(10)])
        flattened = arr.reshape((-1,))
        self.assertEqual((1000,), flattened.shape)
        self.assertNotIsInstance(flattened, ArrayWithUnits)

    def test_can_maintain_array_with_units_when_reshaping_2d(self):
        arr = ArrayWithUnits(
            np.zeros((1, 10)),
            [ContrivedDimension(10), ContrivedDimension2(10)])
        squeezed = arr.reshape((10,))
        self.assertEqual((10,), squeezed.shape)
        self.assertIsInstance(squeezed, ArrayWithUnits)
        self.assertIsInstance(squeezed.dimensions[0], ContrivedDimension2)

    def test_can_maintain_array_with_units_when_squeezing_2d(self):
        arr = ArrayWithUnits(
            np.zeros((1, 10)),
            [ContrivedDimension(10), ContrivedDimension2(10)])
        squeezed = arr.squeeze()
        self.assertEqual((10,), squeezed.shape)
        self.assertIsInstance(squeezed, ArrayWithUnits)
        self.assertIsInstance(squeezed.dimensions[0], ContrivedDimension2)

    def test_can_maintain_array_with_units_when_reshaping_3d(self):
        arr = ArrayWithUnits(
            np.zeros((3, 1, 10)),
            [
                ContrivedDimension(10),
                IdentityDimension(),
                ContrivedDimension2(10)
            ])
        squeezed = arr.reshape((3, 10))
        self.assertEqual((3, 10), squeezed.shape)
        self.assertIsInstance(squeezed, ArrayWithUnits)
        self.assertIsInstance(squeezed.dimensions[0], ContrivedDimension)
        self.assertIsInstance(squeezed.dimensions[1], ContrivedDimension2)

    def test_can_maintain_array_with_units_when_squeezing_3d(self):
        arr = ArrayWithUnits(
            np.zeros((3, 1, 10)),
            [
                ContrivedDimension(10),
                IdentityDimension(),
                ContrivedDimension2(10)
            ])
        squeezed = arr.squeeze()
        self.assertEqual((3, 10), squeezed.shape)
        self.assertIsInstance(squeezed, ArrayWithUnits)
        self.assertIsInstance(squeezed.dimensions[0], ContrivedDimension)
        self.assertIsInstance(squeezed.dimensions[1], ContrivedDimension2)

    def test_squeeze_with_no_single_dimension_is_unchanged(self):
        arr = ArrayWithUnits(
            np.zeros((3, 2, 10)),
            [
                ContrivedDimension(10),
                IdentityDimension(),
                ContrivedDimension2(10)
            ])
        squeezed = arr.squeeze()
        self.assertEqual((3, 2, 10), squeezed.shape)
        self.assertIsInstance(squeezed, ArrayWithUnits)
        self.assertIsInstance(squeezed.dimensions[0], ContrivedDimension)
        self.assertIsInstance(squeezed.dimensions[1], IdentityDimension)
        self.assertIsInstance(squeezed.dimensions[2], ContrivedDimension2)

    def test_assigns_size_where_appropriate(self):
        arr = ArrayWithUnits(
            np.zeros((100, 10)),
            [ContrivedDimension(10), ContrivedDimension2(10)])
        self.assertEqual(10, arr.dimensions[1].size)

    def test_can_create_new_array_from_example(self):
        arr = ArrayWithUnits(
            np.zeros((100, 10)),
            [ContrivedDimension(10), ContrivedDimension2(10)])
        arr2 = ArrayWithUnits.from_example(np.zeros((90, 5)), arr)
        self.assertSequenceEqual(arr.dimensions, arr2.dimensions)
        self.assertEqual((90, 5), arr2.shape)

    def test_zeros(self):
        arr = ArrayWithUnits(
            np.zeros((100, 10)),
            [ContrivedDimension(10), ContrivedDimension2(10)])
        arr2 = ArrayWithUnits.zeros(arr)
        self.assertEqual(arr.shape, arr2.shape)
        self.assertSequenceEqual(arr.dimensions, arr2.dimensions)
        np.testing.assert_allclose(arr2, 0)

    def test_zeros_dtype(self):
        arr = ArrayWithUnits(
            np.zeros((100, 10), dtype=np.complex128),
            [ContrivedDimension(10), ContrivedDimension2(10)])
        arr2 = ArrayWithUnits.zeros(arr)
        self.assertEqual(arr.dtype, arr2.dtype)

    def test_size_is_not_modified_on_example_dimensions(self):
        arr = ArrayWithUnits(
            np.zeros((100, 10)),
            [ContrivedDimension(10), ContrivedDimension2(10)])
        arr2 = ArrayWithUnits.from_example(np.zeros((90, 5)), arr)
        self.assertEqual(100, arr.dimensions[0].size)
        self.assertEqual(10, arr.dimensions[1].size)
        self.assertEqual(90, arr2.dimensions[0].size)
        self.assertEqual(5, arr2.dimensions[1].size)

    def test_can_use_ellipsis_to_get_entire_array(self):
        raw = np.zeros((10, 10, 10))
        dims = (
            ContrivedDimension(10),
            ContrivedDimension(10),
            ContrivedDimension(10)
        )
        arr = ContrivedArray(raw, dims)
        result = arr[...]
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual((10, 10, 10), result.shape)
        self.assertEqual(3, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], ContrivedDimension)
        self.assertIsInstance(result.dimensions[1], ContrivedDimension)
        self.assertIsInstance(result.dimensions[2], ContrivedDimension)

    def test_can_use_ellipsis_to_get_last_two_dimensions(self):
        raw = np.zeros((10, 10, 10))
        dims = (
            ContrivedDimension(10),
            ContrivedDimension(10),
            ContrivedDimension(10)
        )
        arr = ContrivedArray(raw, dims)
        result = arr[ContrivedSlice(10, 30), ...]
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual((2, 10, 10), result.shape)
        self.assertEqual(3, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], ContrivedDimension)
        self.assertIsInstance(result.dimensions[1], ContrivedDimension)
        self.assertIsInstance(result.dimensions[2], ContrivedDimension)

    def test_can_use_ellipsis_to_get_first_two_dimensions(self):
        raw = np.zeros((10, 10, 10))
        dims = (
            ContrivedDimension(10),
            ContrivedDimension(10),
            ContrivedDimension(10)
        )
        arr = ContrivedArray(raw, dims)
        result = arr[..., ContrivedSlice(10, 30)]
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual((10, 10, 2), result.shape)
        self.assertEqual(3, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], ContrivedDimension)
        self.assertIsInstance(result.dimensions[1], ContrivedDimension)
        self.assertIsInstance(result.dimensions[2], ContrivedDimension)

    def test_can_multiply(self):
        raw = np.ones((8, 9))
        arr = ContrivedArray(
            raw, (ContrivedDimension(10), ContrivedDimension2(10)))
        result = arr * 10
        np.testing.assert_allclose(result, 10)

    def test_can_get_single_scalar_from_max_with_no_axis(self):
        raw = np.zeros((8, 9))
        arr = ContrivedArray(
            raw, (ContrivedDimension(10), ContrivedDimension2(10)))
        result = arr.max()
        self.assertIsInstance(result, float)
        self.assertEqual(0, result)

    def test_max_array_maintains_correct_dimensions_axis_0(self):
        raw = np.zeros((8, 9))
        arr = ContrivedArray(
            raw, (ContrivedDimension(10), ContrivedDimension2(10)))
        result = arr.max(axis=0)
        self.assertEqual((9,), result.shape)
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual(1, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], ContrivedDimension2)

    def test_max_array_maintains_correct_dimensions_axis_1(self):
        raw = np.zeros((8, 9))
        arr = ContrivedArray(
            raw, (ContrivedDimension(10), ContrivedDimension2(10)))
        result = arr.max(axis=1)
        self.assertEqual((8,), result.shape)
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual(1, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], ContrivedDimension)

    def test_max_supports_keepdims(self):
        raw = np.random.random_sample((10, 9, 8, 7))
        arr = ArrayWithUnits(
            raw,
            dimensions=[
                ContrivedDimension(10),
                ContrivedDimension2(10),
                ContrivedDimension(10),
                ContrivedDimension2(10)
            ])
        result = arr.max(axis=(1, 2), keepdims=True)
        self.assertEqual((10, 1, 1, 7), result.shape)
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual(4, len(result.dimensions))
        print result.dimensions
        self.assertIsInstance(result.dimensions[0], ContrivedDimension)
        self.assertIsInstance(result.dimensions[1], IdentityDimension)
        self.assertIsInstance(result.dimensions[2], IdentityDimension)
        self.assertIsInstance(result.dimensions[3], ContrivedDimension2)

    def test_max_supports_multiple_axes(self):
        raw = np.random.random_sample((10, 9, 8, 7))
        arr = ArrayWithUnits(
            raw,
            dimensions=[
                ContrivedDimension(10),
                IdentityDimension(),
                IdentityDimension(),
                ContrivedDimension2(10)
            ])
        result = arr.max(axis=(1, 2))
        self.assertEqual((10, 7), result.shape)
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual(2, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], ContrivedDimension)
        self.assertIsInstance(result.dimensions[1], ContrivedDimension2)

    def test_get_single_scalar_from_sum_with_no_axis(self):
        raw = np.zeros((8, 9))
        arr = ContrivedArray(
            raw, (ContrivedDimension(10), ContrivedDimension2(10)))
        result = arr.sum()
        self.assertIsInstance(result, float)
        self.assertEqual(0, result)

    def test_array_maintains_correct_dimension_after_reduction(self):
        raw = np.zeros((8, 9))
        arr = ContrivedArray(
            raw, (ContrivedDimension(10), ContrivedDimension2(10)))
        result = arr.sum(axis=1)
        self.assertEqual((8,), result.shape)
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual(1, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], ContrivedDimension)

    def test_array_maintains_correct_dimension_after_reduction2(self):
        raw = np.zeros((8, 9))
        arr = ContrivedArray(
            raw, (ContrivedDimension(10), ContrivedDimension2(10)))
        result = arr.sum(axis=0)
        self.assertEqual((9,), result.shape)
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual(1, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], ContrivedDimension2)

    def test_array_maintains_correct_dimension_after_reduction3(self):
        raw = np.zeros((8, 9))
        arr = ContrivedArray(
            raw, (ContrivedDimension(10), ContrivedDimension2(10)))
        result = np.sum(arr, axis=0)
        self.assertEqual((9,), result.shape)
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual(1, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], ContrivedDimension2)

    def test_array_cannot_maintain_correct_dimension(self):
        raw = np.zeros((10, 10))
        arr = ContrivedArray(
            raw, (ContrivedDimension(10), ContrivedDimension2(10)))
        result = arr.sum(axis=1)
        self.assertEqual((10,), result.shape)
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual(1, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], ContrivedDimension)

    def test_array_maintains_correct_dimensions_after_dot(self):
        raw = np.zeros((8, 9))
        arr = ContrivedArray(
            raw, (ContrivedDimension(10), ContrivedDimension2(10)))
        result = arr.dot(np.zeros(9))
        self.assertEqual((8,), result.shape)
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual(1, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], ContrivedDimension)

    @unittest2.skip(
        'this test fails because there is no hook to intercept this call')
    def test_array_maintains_correct_dimensions_after_dot2(self):
        raw = np.zeros((8, 9))
        arr = ContrivedArray(
            raw, (ContrivedDimension(10), ContrivedDimension2(10)))
        result = np.dot(arr, np.zeros(9))
        self.assertEqual((8,), result.shape)
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual(1, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], ContrivedDimension)

    def test_can_select_subset_using_boolean_array(self):
        raw = np.arange(10)
        arr = ContrivedArray(raw, (ContrivedDimension(10),))
        result = arr[arr >= 5]
        self.assertIsInstance(result, np.ndarray)
        self.assertEqual((5,), result.shape)

    def test_correct_result_of_indexing_using_boolean_array(self):
        raw = np.random.random_sample((8, 9))
        arr = ContrivedArray(raw, (ContrivedDimension(10), IdentityDimension()))
        result = arr[arr > 0.5]
        self.assertIsInstance(result, np.ndarray)
        self.assertNotIsInstance(result, ArrayWithUnits)

    def test_can_select_subset_using_equality(self):
        raw = np.arange(10)
        arr = ContrivedArray(raw, (ContrivedDimension(10),))
        result = arr[(arr == 5) | (arr == 6)]
        self.assertIsInstance(result, np.ndarray)
        self.assertEqual((2,), result.shape)

    def test_can_set_subset_using_boolean_array(self):
        raw = np.zeros(10)
        arr = ContrivedArray(raw, (ContrivedDimension(10),))
        arr[arr == 0] = 10
        np.testing.assert_allclose(arr, 10)

    def test_can_apply_new_axis(self):
        raw = np.zeros((3, 5))
        arr = ContrivedArray(
            raw, (ContrivedDimension(10), ContrivedDimension2(10)))
        result = arr[None, -1]
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual((1, 5), result.shape)
        self.assertEqual(2, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], IdentityDimension)
        self.assertIsInstance(result.dimensions[1], ContrivedDimension2)

    def test_can_apply_new_axis_at_end(self):
        raw = np.zeros((3, 5))
        arr = ContrivedArray(
            raw, (ContrivedDimension(10), ContrivedDimension2(10)))
        result = arr[-1, None]
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual((1, 5), result.shape)
        self.assertEqual(2, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], IdentityDimension)
        self.assertIsInstance(result.dimensions[1], ContrivedDimension2)

    def test_can_apply_new_axis_in_middle(self):
        raw = np.zeros((3, 5))
        arr = ContrivedArray(
            raw, (ContrivedDimension(10), ContrivedDimension2(10)))
        result = arr[:, None, :]
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual((3, 1, 5), result.shape)
        self.assertEqual(3, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], ContrivedDimension)
        self.assertIsInstance(result.dimensions[1], IdentityDimension)
        self.assertIsInstance(result.dimensions[2], ContrivedDimension2)

    def test_can_apply_two_new_axes(self):
        raw = np.zeros((3, 5))
        arr = ContrivedArray(
            raw, (ContrivedDimension(10), ContrivedDimension2(10)))
        result = arr[None, None, -1]
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual((1, 1, 5), result.shape)
        self.assertEqual(3, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], IdentityDimension)
        self.assertIsInstance(result.dimensions[1], IdentityDimension)
        self.assertIsInstance(result.dimensions[2], ContrivedDimension2)

    def test_correct_axis_is_preserved(self):
        raw = np.zeros((10, 10))
        arr = ContrivedArray(
            raw, (ContrivedDimension(10), ContrivedDimension2(10)))
        result = arr[0]
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual(1, len(result.dimensions))
        self.assertEqual((10,), result.shape)
        self.assertIsInstance(result.dimensions[0], ContrivedDimension2)

    def test_multiple_axis_types(self):
        raw = np.zeros((10, 10))
        arr = ContrivedArray(
            raw, (ContrivedDimension(10), ContrivedDimension2(10)))
        result = arr[ContrivedSlice(20, 40), ContrivedSlice(20, 50)]
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual(2, len(result.dimensions))
        self.assertEqual((2, 3), result.shape)
        self.assertIsInstance(result.dimensions[0], ContrivedDimension)
        self.assertIsInstance(result.dimensions[1], ContrivedDimension2)

    def test_can_use_list_of_integers_as_index(self):
        raw = np.zeros((10, 10))
        arr = ContrivedArray(
            raw, (ContrivedDimension(10), ContrivedDimension2(10)))
        result = arr[[1, 3, 5]]
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual((3, 10), result.shape)
        self.assertEqual(2, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], IdentityDimension)
        self.assertIsInstance(result.dimensions[1], ContrivedDimension2)

    def test_can_use_multiple_integers(self):
        raw = np.zeros((10, 10))
        arr = ContrivedArray(
            raw, (ContrivedDimension(10), ContrivedDimension2(10)))
        result = arr[0, 0]
        self.assertEqual(0, result)

    def test_1d(self):
        raw = np.zeros(10)
        arr = ContrivedArray(raw, (ContrivedDimension(10),))
        result = arr[1:4]
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual((3,), result.shape)
        self.assertIsInstance(result.dimensions[0], ContrivedDimension)

    def test_can_get_modified_dimension(self):
        raw = np.zeros(10)
        arr = ContrivedArray(
            raw, (AsciiCharacterDimension(ascii_lowercase[:10]),))
        result = arr[1:4]
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual((3,), result.shape)
        self.assertIsInstance(result.dimensions[0], AsciiCharacterDimension)
        self.assertEqual(ascii_lowercase[1:4], result.dimensions[0].labels)

    def test_can_get_custom_slice(self):
        raw = np.zeros((10, 10))
        arr = ContrivedArray(raw, (None, ContrivedDimension(10)))
        custom_slice = ContrivedSlice(50, 70)
        result = arr[5:7, custom_slice]
        self.assertEqual((2, 2), result.shape)

    def test_can_set_custom_slice(self):
        raw = np.zeros((10, 10))
        arr = ContrivedArray(raw, (None, ContrivedDimension(10)))
        custom_slice = ContrivedSlice(50, 70)
        arr[5:7, custom_slice] = 1
        np.testing.assert_allclose(arr[5:7, 5:7], 1)

    def test_can_set_from_list_of_integers(self):
        raw = np.zeros((10, 10))
        arr = ContrivedArray(
            raw, (ContrivedDimension(10), ContrivedDimension2(10)))
        arr[[1, 3, 5]] = 5
        np.testing.assert_allclose(arr[[1, 3, 5]], 5)

    def test_dimensions_must_match(self):
        raw = np.zeros((3, 3, 3))
        self.assertRaises(
            ValueError,
            lambda: ContrivedArray(raw, (None, ContrivedDimension(5))))

    def test_custom_dimensions_are_preserved_after_slice(self):
        raw = np.zeros((10, 10))
        arr = ContrivedArray(raw, (None, ContrivedDimension(10)))
        custom_slice = ContrivedSlice(50, 70)
        result = arr[5:7, custom_slice]
        self.assertIsInstance(result.dimensions[0], IdentityDimension)
        self.assertEqual(ContrivedDimension(10), result.dimensions[1])

    def test_custom_dimensions_are_preserved_after_slice_that_removes_dimension(
            self):
        raw = np.zeros((10, 10))
        arr = ContrivedArray(raw, (None, ContrivedDimension(10)))
        custom_slice = ContrivedSlice(50, 70)
        result = arr[0, custom_slice]
        self.assertEqual(1, result.ndim)
        self.assertEqual((2,), result.shape)
        self.assertEqual(ContrivedDimension(10), result.dimensions[0])

    def test_can_get_single_value_from_custom_dimension(self):
        raw = np.zeros(10)
        arr = ContrivedArray(raw, (ContrivedDimension(10),))
        result = arr[0]
        self.assertEqual(0, result)
        self.assertIsInstance(result, float)

    def test_can_set_single_value_in_custom_dimension(self):
        raw = np.zeros(10)
        arr = ContrivedArray(raw, (ContrivedDimension(10),))
        arr[0] = 1
        self.assertEqual(1, arr[0])

    def test_sliding_window_maintains_dtype(self):
        raw = np.zeros(10, dtype=np.uint8)
        arr = ContrivedArray(raw, (ContrivedDimension(10),))
        new_arr = arr.sliding_window((ContrivedSlice(0, 20),))
        self.assertEqual(np.uint8, new_arr.dtype)

    def test_custom_dimension_is_maintained_for_sliding_window_1d(self):
        raw = np.zeros(10)
        arr = ContrivedArray(raw, (ContrivedDimension(10),))
        new_arr = arr.sliding_window((ContrivedSlice(0, 20),))
        self.assertIsInstance(new_arr, ArrayWithUnits)
        self.assertEqual((5, 2), new_arr.shape)
        self.assertIsInstance(new_arr.dimensions[0], ContrivedDimension)
        self.assertIsInstance(new_arr.dimensions[1], ContrivedDimension)
        self.assertEqual(20, new_arr.dimensions[0].factor)
        self.assertEqual(10, new_arr.dimensions[1].factor)

    def test_item_access_is_correct_after_sliding_window_1d(self):
        raw = np.zeros(10)
        arr = ContrivedArray(raw, (ContrivedDimension(10),))
        new_arr = arr.sliding_window((ContrivedSlice(0, 20),))
        sliced = new_arr[ContrivedSlice(0, 40)]
        self.assertEqual((2, 2), sliced.shape)

    def test_custom_dimension_is_maintained_for_sw_1d_with_step(self):
        raw = np.zeros(10)
        arr = ContrivedArray(raw, (ContrivedDimension(10),))
        new_arr = arr.sliding_window(
            (ContrivedSlice(0, 20),), (ContrivedSlice(0, 10),))
        self.assertIsInstance(new_arr, ArrayWithUnits)
        self.assertEqual((9, 2), new_arr.shape)
        self.assertIsInstance(new_arr.dimensions[0], ContrivedDimension)
        self.assertIsInstance(new_arr.dimensions[1], ContrivedDimension)
        self.assertEqual(20, new_arr.dimensions[0].factor)
        self.assertEqual(10, new_arr.dimensions[1].factor)

    def test_custom_dimension_is_maintained_for_sliding_window(self):
        raw = np.zeros((10, 10))
        arr = ContrivedArray(raw, (ContrivedDimension(10), None))
        new_arr = arr.sliding_window((ContrivedSlice(0, 20), 10))
        self.assertIsInstance(new_arr, ArrayWithUnits)
        self.assertEqual((5, 2, 10), new_arr.shape)
        self.assertIsInstance(new_arr.dimensions[0], ContrivedDimension)
        self.assertIsInstance(new_arr.dimensions[1], ContrivedDimension)
        self.assertIsInstance(new_arr.dimensions[2], IdentityDimension)
        self.assertEqual(20, new_arr.dimensions[0].factor)
        self.assertEqual(10, new_arr.dimensions[1].factor)

    def test_custom_dimension_is_maintained_for_sw_with_step(self):
        raw = np.zeros((10, 10))
        arr = ContrivedArray(raw, (ContrivedDimension(10), None))
        new_arr = arr.sliding_window(
            (ContrivedSlice(0, 20), 10), (ContrivedSlice(0, 10), 10))
        self.assertIsInstance(new_arr, ArrayWithUnits)
        self.assertEqual((9, 2, 10), new_arr.shape)
        self.assertIsInstance(new_arr.dimensions[0], ContrivedDimension)
        self.assertIsInstance(new_arr.dimensions[1], ContrivedDimension)
        self.assertIsInstance(new_arr.dimensions[2], IdentityDimension)
        self.assertEqual(20, new_arr.dimensions[0].factor)
        self.assertEqual(10, new_arr.dimensions[1].factor)

    def test_custom_dimension_is_not_maintained_for_sliding_window(self):
        raw = np.zeros((10, 10))
        arr = ContrivedArray(raw, (ContrivedDimension(10), None))
        new_arr = arr.sliding_window((ContrivedSlice(0, 20), 2))
        self.assertIsInstance(new_arr, ArrayWithUnits)
        self.assertEqual((25, 2, 2), new_arr.shape)
        self.assertIsInstance(new_arr.dimensions[0], IdentityDimension)
        self.assertIsInstance(new_arr.dimensions[1], IdentityDimension)
        self.assertIsInstance(new_arr.dimensions[2], IdentityDimension)

    def test_custom_dimension_is_not_maintained_for_sw_with_step(self):
        raw = np.zeros((10, 10))
        arr = ContrivedArray(raw, (ContrivedDimension(10), None))
        new_arr = arr.sliding_window(
            (ContrivedSlice(0, 20), 2), (ContrivedSlice(0, 10), 1))
        self.assertIsInstance(new_arr, ArrayWithUnits)
        self.assertEqual((81, 2, 2), new_arr.shape)
        self.assertIsInstance(new_arr.dimensions[0], IdentityDimension)
        self.assertIsInstance(new_arr.dimensions[1], IdentityDimension)
        self.assertIsInstance(new_arr.dimensions[2], IdentityDimension)



import unittest2
from zounds.core import ArrayWithUnits
from zounds.timeseries import TimeDimension, Seconds, VariableRateTimeSeries
from onset import BasePeakPicker
import numpy as np


class BasePeakPickerTests(unittest2.TestCase):

    class PeakPicker(BasePeakPicker):
        def _onset_indices(self, data):
            indices = np.random.permutation(np.arange(len(data)))[:3]
            indices.sort()
            return indices

    def test_peak_picker_returns_variable_rate_time_series(self):
        data = ArrayWithUnits(
            np.zeros(100),
            dimensions=[TimeDimension(frequency=Seconds(1))])
        picker = BasePeakPickerTests.PeakPicker()
        results = picker._process(data).next()
        self.assertEqual(3, len(results))
        self.assertIsInstance(results, VariableRateTimeSeries)



from zounds.core import ArrayWithUnits
from zounds.persistence import DimensionEncoder, DimensionDecoder
from featureflow import Node, Decoder, Feature, NumpyMetaData
import struct
import json
import numpy as np


def _np_from_buffer(b, shape, dtype):
    try:
        shape = tuple(int(x) for x in shape)
    except TypeError:
        shape = int(shape)
    f = np.frombuffer if len(b) else np.fromstring
    return f(b, dtype=dtype).reshape(shape)


class ArrayWithUnitsEncoder(Node):
    content_type = 'application/octet-stream'

    def __init__(self, needs=None):
        super(ArrayWithUnitsEncoder, self).__init__(needs=needs)
        self.encoder = DimensionEncoder()
        self.dimensions = None
        self.nmpy = None

    def _process(self, data):
        if self.dimensions is None:
            self.dimensions = data.dimensions
            d = list(self.encoder.encode(self.dimensions))
            encoded = json.dumps(d)
            yield struct.pack('I', len(encoded))
            yield encoded
        if self.nmpy is None:
            self.nmpy = NumpyMetaData(data.dtype, data.shape[1:])
            yield self.nmpy.pack()

        yield data.tostring()


class PackedArrayWithUnitsEncoder(Node):
    content_type = 'application/octet-stream'

    def __init__(self, needs=None):
        super(PackedArrayWithUnitsEncoder, self).__init__(needs=needs)
        self.encoder = DimensionEncoder()
        self.dimensions = None
        self.nmpy = None

    def _process(self, data):
        if self.dimensions is None:
            self.dimensions = data.dimensions
            d = list(self.encoder.encode(self.dimensions))
            encoded = json.dumps(d)
            yield struct.pack('I', len(encoded))
            yield encoded

        packed = np.packbits(data.astype(np.uint8), axis=-1)

        if self.nmpy is None:
            self.nmpy = NumpyMetaData(packed.dtype, packed.shape[1:])
            yield self.nmpy.pack()

        yield packed.tostring()


class ArrayWithUnitsDecoder(Decoder):
    def __init__(self):
        super(ArrayWithUnitsDecoder, self).__init__()

    def __call__(self, flo):
        nbytes = struct.calcsize('I')
        json_len = struct.unpack('I', flo.read(nbytes))[0]
        d = json.loads(flo.read(json_len))
        decoder = DimensionDecoder()
        dimensions = list(decoder.decode(d))

        metadata, bytes_read = NumpyMetaData.unpack(flo)
        leftovers = flo.read()
        leftover_bytes = len(leftovers)
        first_dim = leftover_bytes / metadata.totalsize
        dim = (first_dim,) + metadata.shape
        raw = _np_from_buffer(leftovers, dim, metadata.dtype)

        return ArrayWithUnits(raw, dimensions)

    def __iter__(self, flo):
        yield self(flo)


class ArrayWithUnitsFeature(Feature):
    def __init__(
            self,
            extractor,
            needs=None,
            store=False,
            key=None,
            encoder=ArrayWithUnitsEncoder,
            decoder=ArrayWithUnitsDecoder(),
            **extractor_args):
        super(ArrayWithUnitsFeature, self).__init__(
                extractor,
                needs=needs,
                store=store,
                encoder=encoder,
                decoder=decoder,
                key=key,
                **extractor_args)


class PackedArrayWithUnitsFeature(Feature):
    def __init__(
            self,
            extractor,
            needs=None,
            store=False,
            key=None,
            encoder=PackedArrayWithUnitsEncoder,
            decoder=ArrayWithUnitsDecoder(),
            **extractor_args):
        super(ArrayWithUnitsFeature, self).__init__(
                extractor,
                needs=needs,
                store=store,
                encoder=encoder,
                decoder=decoder,
                key=key,
                **extractor_args)



import unittest2
import featureflow as ff
import numpy as np

from zounds.util import simple_in_memory_settings
from preprocess import Slicer, PreprocessingPipeline

from zounds.core import IdentityDimension, ArrayWithUnits
from zounds.spectral import FrequencyBand, LinearScale, FrequencyDimension
from zounds.timeseries import TimeDimension, Seconds


class SlicerTests(unittest2.TestCase):
    def get_model(self, slicex):
        @simple_in_memory_settings
        class Model(ff.BaseModel):
            sliced = ff.PickleFeature(
                    Slicer,
                    slicex=slicex,
                    store=False)

            pipeline = ff.PickleFeature(
                    PreprocessingPipeline,
                    needs=(sliced,),
                    store=True)

        return Model

    def test_inversion_maintains_data_type(self):
        training = np.ones((100, 30), dtype=np.complex128)
        Model = self.get_model(slicex=slice(10, 20))
        _id = Model.process(sliced=training)
        model = Model(_id)
        data = np.ones((100, 30), dtype=np.complex128)
        transformed = model.pipeline.transform(data)
        self.assertEqual((100, 10), transformed.data.shape)
        inverted = transformed.inverse_transform()
        self.assertEqual(np.complex128, inverted.dtype)

    def test_can_slice_array_with_slice_instance(self):
        training = np.ones((100, 30))
        Model = self.get_model(slicex=slice(10, 20))
        _id = Model.process(sliced=training)
        model = Model(_id)
        data = np.ones((100, 30))
        transformed = model.pipeline.transform(data)
        self.assertEqual((100, 10), transformed.data.shape)
        inverted = transformed.inverse_transform()
        self.assertEqual((100, 30), inverted.shape)
        np.testing.assert_allclose(transformed.data, 1)
        np.testing.assert_allclose(inverted[:, :10], 0)
        np.testing.assert_allclose(inverted[:, 20:], 0)

    def test_can_slice_multi_dimensional(self):
        training = np.ones((100, 30, 50))
        Model = self.get_model(slicex=slice(10, 20))
        _id = Model.process(sliced=training)
        model = Model(_id)
        data = np.ones((100, 30, 50))
        transformed = model.pipeline.transform(data)
        self.assertEqual((100, 30, 10), transformed.data.shape)
        inverted = transformed.inverse_transform()
        self.assertEqual((100, 30, 50), inverted.shape)
        np.testing.assert_allclose(transformed.data, 1)
        np.testing.assert_allclose(inverted[..., :10], 0)
        np.testing.assert_allclose(inverted[..., 20:], 0)

    def test_can_invert_array_with_units(self):
        td = TimeDimension(Seconds(1))
        fd = FrequencyDimension(LinearScale(FrequencyBand(0, 20000), 100))
        dimensions = [IdentityDimension(), td, fd]
        training = ArrayWithUnits(np.zeros((10, 5, 100)), dimensions)
        Model = self.get_model(slicex=FrequencyBand(1000, 10000))
        _id = Model.process(sliced=training)
        model = Model(_id)
        data = ArrayWithUnits(np.ones((2, 5, 100)), dimensions)
        transformed = model.pipeline.transform(data)
        inverted = transformed.inverse_transform()
        self.assertEqual((2, 5, 100), inverted.shape)
        self.assertEqual(IdentityDimension(), inverted.dimensions[0])
        self.assertEqual(td, inverted.dimensions[1])
        self.assertEqual(fd, inverted.dimensions[2])



import unittest2
import featureflow as ff
from preprocess import Weighted, InstanceScaling, PreprocessingPipeline
from zounds.spectral import \
    AWeighting, GeometricScale, FrequencyDimension, FrequencyAdaptive
from zounds.core import ArrayWithUnits
from zounds.timeseries import TimeDimension, Seconds
from zounds.util import simple_in_memory_settings
import numpy as np


class WeightedTests(unittest2.TestCase):

    def test_can_apply_weighting(self):

        @simple_in_memory_settings
        class Model(ff.BaseModel):
            weighted = ff.PickleFeature(
                Weighted,
                weighting=AWeighting(),
                store=False)

            scaled = ff.PickleFeature(
                InstanceScaling,
                needs=weighted,
                store=False)

            pipeline = ff.PickleFeature(
                PreprocessingPipeline,
                needs=(weighted, scaled),
                store=True)

        training = ArrayWithUnits(
            np.random.random_sample((100, 15)),
            dimensions=[
                TimeDimension(Seconds(1)),
                FrequencyDimension(GeometricScale(100, 1000, 0.1, 15))
            ]
        )

        _id = Model.process(weighted=training)
        model = Model(_id)

        test = ArrayWithUnits(
            np.random.random_sample((10, 15)),
            dimensions=[
                TimeDimension(Seconds(1)),
                FrequencyDimension(GeometricScale(100, 1000, 0.1, 15))
            ]
        )

        result = model.pipeline.transform(test)
        inverted = result.inverse_transform()

        np.testing.assert_allclose(test, inverted)

    def test_can_apply_weighting_to_frequency_adaptive_transform(self):
        @simple_in_memory_settings
        class Model(ff.BaseModel):
            weighted = ff.PickleFeature(
                Weighted,
                weighting=AWeighting(),
                store=False)

            scaled = ff.PickleFeature(
                InstanceScaling,
                needs=weighted,
                store=False)

            pipeline = ff.PickleFeature(
                PreprocessingPipeline,
                needs=(weighted, scaled),
                store=True)

        scale = GeometricScale(100, 1000, 0.1, 15)

        training = FrequencyAdaptive(
            [np.ones((100, x)) for x in xrange(1, len(scale) + 1)],
            time_dimension=TimeDimension(Seconds(1)),
            scale=scale)

        _id = Model.process(weighted=training)
        model = Model(_id)

        test = FrequencyAdaptive(
            [np.ones((10, x)) for x in xrange(1, len(scale) + 1)],
            time_dimension=TimeDimension(Seconds(1)),
            scale=scale)

        result = model.pipeline.transform(test)
        inverted = result.inverse_transform()

        self.assertIsInstance(inverted, FrequencyAdaptive)
        np.testing.assert_allclose(test, inverted)





from trainer import Trainer
import numpy as np


class SupervisedTrainer(Trainer):
    def __init__(
            self,
            model,
            loss,
            optimizer,
            epochs,
            batch_size,
            holdout_percent=0.0,
            data_preprocessor=lambda x: x,
            label_preprocessor=lambda x: x):

        super(SupervisedTrainer, self).__init__(
            epochs,
            batch_size)

        self.label_preprocessor = label_preprocessor
        self.data_preprocessor = data_preprocessor
        self.holdout_percent = holdout_percent
        self.optimizer = optimizer(model)
        self.loss = loss
        self.model = model

    def train(self, data):
        import torch
        from torch.autograd import Variable

        model = self.model.cuda()
        loss = self.loss.cuda()

        data, labels = data['data'], data['labels']

        test_size = int(self.holdout_percent * len(data))
        test_data, test_labels = data[:test_size], labels[:test_size]
        data, labels = data[test_size:], labels[test_size:]

        if data is labels:
            # this is an autoencoder scenario, so let's saved on memory
            data = data.astype(np.float32)
            test_data = data.astype(np.float32)
            labels = data
            test_labels = labels
        else:
            data = data.astype(np.float32)
            test_data = test_data.astype(np.float32)
            labels = labels.astype(np.float32)
            test_labels = test_labels.astype(np.float32)

        def batch(d, l, test=False):
            d = self.data_preprocessor(d).astype(np.float32)
            l = self.label_preprocessor(l).astype(np.float32)
            inp = torch.from_numpy(d)
            inp = inp.cuda()
            inp_v = Variable(inp, volatile=test)
            output = model(inp_v)

            labels_t = torch.from_numpy(l)
            labels_t = labels_t.cuda()
            labels_v = Variable(labels_t)

            error = loss(output, labels_v)

            if not test:
                error.backward()
                self.optimizer.step()

            return error.data[0]

        for epoch in xrange(self.epochs):
            for i in xrange(0, len(data), self.batch_size):

                model.zero_grad()

                # training batch
                minibatch_slice = slice(i, i + self.batch_size)
                minibatch_data = data[minibatch_slice]
                minibatch_labels = labels[minibatch_slice]

                e = training_error = batch(
                    minibatch_data, minibatch_labels, test=False)

                # test batch
                if test_size:
                    indices = np.random.randint(0, test_size, self.batch_size)
                    test_batch_data = test_data[indices, ...]
                    test_batch_labels = test_labels[indices, ...]

                    te = test_error = batch(
                        test_batch_data, test_batch_labels, test=True)
                else:
                    te = 'n/a'

                if i % 10 == 0:
                    print 'Epoch {epoch}, batch {i}, train error {e}, test error {te}'.format(
                        **locals())

        return model



import numpy as np
from timeseries import TimeSlice
from duration import Seconds, Picoseconds
from featureflow import Feature, BaseNumpyDecoder, NumpyEncoder
from itertools import izip, chain


class VariableRateTimeSeries(object):
    def __init__(self, data):
        super(VariableRateTimeSeries, self).__init__()
        if isinstance(data, np.recarray):
            self._data = data
            return

        data = sorted(list(data), key=lambda x: x[0])
        try:
            example = data[0][1]
            shape = example.shape
            dtype = example.dtype
        except IndexError:
            shape = (0,)
            dtype = np.uint8
        self._data = np.recarray(len(data), dtype=[
            ('timeslice', TimeSlice),
            ('slicedata', dtype, shape)])
        self._data[:] = data

    def __len__(self):
        return self._data.__len__()

    def concat(self, other):
        return VariableRateTimeSeries(
            chain(self.iter_slices(), other.iter_slices()))

    def iter_slices(self):
        return izip(self.slices, self.slicedata)

    @property
    def slices(self):
        return self._data.timeslice

    @property
    def slicedata(self):
        return self._data.slicedata

    @property
    def raw_data(self):
        return self._data

    @property
    def span(self):
        try:
            start = self._data.timeslice[0].start
            return TimeSlice(start=start, duration=self.end - start)
        except IndexError:
            return TimeSlice(duration=Seconds(0))

    @property
    def end(self):
        try:
            return self._data.timeslice[-1].end
        except IndexError:
            return Seconds(0)

    def __getitem__(self, index):
        if isinstance(index, TimeSlice):
            # TODO: Consider using a bisection approach here to make this much
            # faster than this brute-force, O(n) approach
            # compare the beginning of the index to the _end_ of each sample
            # compare the end of the index to the beginning of each sample
            g = ((x.timeslice, x.slicedata) for x in self._data
                 if x.timeslice.end > index.start and
                 (index.duration is None or x.timeslice.start < index.end))
            return VariableRateTimeSeries(g)
        elif isinstance(index, int):
            return self._data[index]
        else:
            return VariableRateTimeSeries(self._data[index])


class VariableRateTimeSeriesEncoder(NumpyEncoder):
    def __init__(self, needs=None):
        super(VariableRateTimeSeriesEncoder, self).__init__(needs=needs)

    def _prepare_data(self, data):
        output = np.recarray(len(data), dtype=[
            ('start', np.int64),
            ('duration', np.int64),
            ('slicedata', data.slicedata.dtype, data.slicedata.shape[1:])
        ])
        ps = Picoseconds(1)
        output.slicedata[:] = data.slicedata
        output.start[:] = [ts.start / ps for ts in data.slices]
        output.duration[:] = [ts.duration / ps for ts in data.slices]
        return output


class VariableRateTimeSeriesDecoder(BaseNumpyDecoder):
    def __init__(self):
        super(VariableRateTimeSeriesDecoder, self).__init__()

    def _gen(self, raw):
        for r in raw:
            ts = TimeSlice(
                    start=Picoseconds(r.start),
                    duration=Picoseconds(r.duration))
            yield (ts, r.slicedata)

    def _wrap_array(self, raw, metadata):
        return VariableRateTimeSeries(self._gen(raw))


class VariableRateTimeSeriesFeature(Feature):
    def __init__(
            self,
            extractor,
            needs=None,
            store=False,
            key=None,
            encoder=VariableRateTimeSeriesEncoder,
            decoder=VariableRateTimeSeriesDecoder(),
            **extractor_args):
        super(VariableRateTimeSeriesFeature, self).__init__(
                extractor,
                needs=needs,
                store=store,
                encoder=encoder,
                decoder=decoder,
                key=key,
                **extractor_args)



import unittest2
from functional import categorical, inverse_categorical
from zounds.core import ArrayWithUnits
from zounds.synthesize import SineSynthesizer
from zounds.timeseries import Seconds, SR11025, TimeSlice
import numpy as np


class CategoricalTests(unittest2.TestCase):
    def test_can_convert_to_categorical_distribution(self):
        samplerate = SR11025()
        synth = SineSynthesizer(samplerate)
        samples = synth.synthesize(Seconds(4), [220, 440, 880])
        _, windowed = samples.sliding_window_with_leftovers(
            TimeSlice(duration=samplerate.frequency * 512),
            TimeSlice(duration=samplerate.frequency * 256))
        c = categorical(windowed, mu=255)
        self.assertEqual(windowed.shape + (255 + 1,), c.shape)
        np.testing.assert_allclose(c.sum(axis=-1), 1)

    def test_can_invert_categorical_distribution(self):
        samplerate = SR11025()
        synth = SineSynthesizer(samplerate)
        samples = synth.synthesize(Seconds(4), [220, 440, 880])
        _, windowed = samples.sliding_window_with_leftovers(
            TimeSlice(duration=samplerate.frequency * 512),
            TimeSlice(duration=samplerate.frequency * 256))
        c = categorical(windowed, mu=255)
        inverted = inverse_categorical(c, mu=255)
        self.assertEqual(windowed.shape, inverted.shape)
        self.assertIsInstance(inverted, ArrayWithUnits)
        self.assertSequenceEqual(windowed.dimensions, inverted.dimensions)




import unittest2
from constantrate import ConstantRateTimeSeries
from zounds.core import ArrayWithUnits, IdentityDimension
from zounds.timeseries import TimeDimension, Seconds, Milliseconds, TimeSlice
import numpy as np


class ConstantRateTimeSeriesTests(unittest2.TestCase):

    def test_raises_when_not_array_with_units_instance(self):
        arr = np.zeros(10)
        self.assertRaises(ValueError, lambda: ConstantRateTimeSeries(arr))

    def test_raises_when_first_dimension_is_not_time_dimension(self):
        raw = np.zeros((10, 3))
        arr = ArrayWithUnits(raw, dimensions=[
            IdentityDimension(), TimeDimension(frequency=Seconds(1))])
        self.assertRaises(ValueError, lambda: ConstantRateTimeSeries(arr))

    def test_iter_slices_yields_evenly_spaced_time_slices(self):
        raw = np.random.random_sample((10, 3))
        arr = ArrayWithUnits(raw, dimensions=[
            TimeDimension(frequency=Milliseconds(500), duration=Seconds(1)),
            IdentityDimension()
        ])
        crts = ConstantRateTimeSeries(arr)
        slices = list(crts.iter_slices())
        self.assertEqual(10, len(slices))

        ts1, d1 = slices[0]
        self.assertEqual(
            TimeSlice(start=Seconds(0), duration=Seconds(1)), ts1)
        np.testing.assert_allclose(raw[0], d1)

        ts2, d2 = slices[1]
        self.assertEqual(
            TimeSlice(start=Milliseconds(500), duration=Seconds(1)), ts2)
        np.testing.assert_allclose(raw[1], d2)



from basedimension import BaseDimensionEncoder, BaseDimensionDecoder
from zounds.spectral import \
    FrequencyBand, FrequencyDimension, LinearScale, GeometricScale, \
    ExplicitScale, ExplicitFrequencyDimension


class ScaleEncoderDecoder(object):
    def __init__(self):
        super(ScaleEncoderDecoder, self).__init__()

    def can_encode(self, scale):
        raise NotImplementedError()

    def can_decode(self, d):
        raise NotImplementedError()

    def decode(self, d):
        raise NotImplementedError()

    def encode(self, scale):
        raise NotImplementedError()


class LinearScaleEncoderDecoder(ScaleEncoderDecoder):
    def __init__(self):
        super(LinearScaleEncoderDecoder, self).__init__()

    def can_encode(self, scale):
        return isinstance(scale, LinearScale)

    def can_decode(self, d):
        return d['name'] == LinearScale.__name__

    def encode(self, scale):
        return dict(
            start_hz=scale.frequency_band.start_hz,
            stop_hz=scale.frequency_band.stop_hz,
            n_bands=scale.n_bands,
            name=scale.__class__.__name__,
            always_even=scale.always_even)

    def decode(self, d):
        band = FrequencyBand(d['start_hz'], d['stop_hz'])
        return LinearScale(band, d['n_bands'], always_even=d['always_even'])


class GeometricScaleEncoderDecoder(ScaleEncoderDecoder):
    def __init__(self):
        super(GeometricScaleEncoderDecoder, self).__init__()

    def can_encode(self, scale):
        return isinstance(scale, GeometricScale)

    def can_decode(self, d):
        return d['name'] == GeometricScale.__name__

    def encode(self, scale):
        return dict(
            start_center_hz=scale.start_center_hz,
            stop_center_hz=scale.stop_center_hz,
            bandwidth_ratio=scale.bandwidth_ratio,
            n_bands=scale.n_bands,
            name=GeometricScale.__name__,
            always_even=scale.always_even)

    def decode(self, d):
        return GeometricScale(
            d['start_center_hz'],
            d['stop_center_hz'],
            d['bandwidth_ratio'],
            d['n_bands'],
            d['always_even'])


class ExplicitScaleEncoderDecoder(ScaleEncoderDecoder):
    def __init__(self):
        super(ExplicitScaleEncoderDecoder, self).__init__()

    def can_encode(self, scale):
        return isinstance(scale, ExplicitScale)

    def can_decode(self, d):
        return d['name'] == ExplicitScale.__name__

    def encode(self, scale):
        bands = [(b.start_hz, b.stop_hz) for b in scale]
        return dict(bands=bands, name=ExplicitScale.__name__)

    def decode(self, d):
        bands = [FrequencyBand(*b) for b in d['bands']]
        return ExplicitScale(bands)


class CompositeScaleEncoderDecoder(object):
    strategies = [
        LinearScaleEncoderDecoder(),
        GeometricScaleEncoderDecoder(),
        ExplicitScaleEncoderDecoder()
    ]

    def __init__(self):
        super(CompositeScaleEncoderDecoder, self).__init__()

    def _find_encoder(self, scale):
        try:
            return next(s for s in self.strategies if s.can_encode(scale))
        except StopIteration:
            raise NotImplementedError('No suitable encoder found')

    def _find_decoder(self, d):
        try:
            return next(s for s in self.strategies if s.can_decode(d))
        except StopIteration:
            raise NotImplementedError('No suitable decoder found')

    def encode(self, scale):
        return self._find_encoder(scale).encode(scale)

    def decode(self, d):
        return self._find_decoder(d).decode(d)


class FrequencyDimensionEncoder(BaseDimensionEncoder):
    def __init__(self):
        super(FrequencyDimensionEncoder, self).__init__(FrequencyDimension)
        self.scale_encoder = CompositeScaleEncoderDecoder()

    def dict(self, freq_dim):
        return self.scale_encoder.encode(freq_dim.scale)


class FrequencyDimensionDecoder(BaseDimensionDecoder):
    def __init__(self):
        super(FrequencyDimensionDecoder, self).__init__(FrequencyDimension)
        self.scale_decoder = CompositeScaleEncoderDecoder()

    def args(self, d):
        return self.scale_decoder.decode(d),


class ExplicitFrequencyDimensionEncoder(BaseDimensionEncoder):
    def __init__(self):
        super(ExplicitFrequencyDimensionEncoder, self).__init__(
            ExplicitFrequencyDimension)
        self.scale_encoder = CompositeScaleEncoderDecoder()

    def dict(self, freq_dim):
        d = self.scale_encoder.encode(freq_dim.scale)
        slices = [(s.start, s.stop) for s in freq_dim.slices]
        d.update(slices=slices)
        return d


class ExplicitFrequencyDimensionDecoder(BaseDimensionDecoder):
    def __init__(self):
        super(ExplicitFrequencyDimensionDecoder, self).__init__(
            ExplicitFrequencyDimension)
        self.scale_decoder = CompositeScaleEncoderDecoder()

    def args(self, d):
        scale = self.scale_decoder.decode(d)
        slices = [slice(start, stop) for (start, stop) in d['slices']]
        return scale, slices



import unittest2
from audiograph import resampled, stft, frequency_adaptive
from zounds.timeseries.samplerate import SR11025, SR22050, SampleRate
from zounds.timeseries.duration import Seconds, Milliseconds
from zounds.util.persistence import simple_in_memory_settings
from zounds.persistence import ArrayWithUnitsFeature
from zounds.synthesize.synthesize import NoiseSynthesizer, SineSynthesizer
from zounds.spectral import GeometricScale, FrequencyAdaptive
import zipfile
from io import BytesIO
import featureflow
import numpy as np


class FrequencyAdaptiveTests(unittest2.TestCase):
    def test_can_compute_frequency_adaptive_feature(self):
        scale = GeometricScale(
            start_center_hz=50,
            stop_center_hz=5000,
            bandwidth_ratio=0.07123,
            n_bands=128)

        fa = frequency_adaptive(
            SampleRate(frequency=Milliseconds(358), duration=Milliseconds(716)),
            scale,
            store_freq_adaptive=True)

        @simple_in_memory_settings
        class Document(fa):
            rasterized = ArrayWithUnitsFeature(
                lambda fa: fa.rasterize(64).astype(np.float32),
                needs=fa.freq_adaptive,
                store=True)

        synth = SineSynthesizer(SR22050())
        samples = synth.synthesize(Seconds(10))
        _id = Document.process(meta=samples.encode())
        doc = Document(_id)
        self.assertIsInstance(doc.freq_adaptive, FrequencyAdaptive)
        self.assertEqual((64, 128), doc.rasterized.shape[-2:])


class ResampledTests(unittest2.TestCase):
    def test_audio_is_resampled(self):
        orig_sample_rate = SR22050()
        new_sample_rate = SR11025()

        rs = resampled(resample_to=new_sample_rate, store_resampled=True)

        @simple_in_memory_settings
        class Document(rs):
            pass

        synth = NoiseSynthesizer(orig_sample_rate)
        samples = synth.synthesize(Seconds(3))
        _id = Document.process(meta=samples.encode())
        doc = Document(_id)
        self.assertEqual(new_sample_rate, doc.resampled.samplerate)
        self.assertEqual(len(samples) // 2, len(doc.resampled))


class StftTests(unittest2.TestCase):
    def test_can_hadle_zip_file(self):
        STFT = stft(resample_to=SR11025(), store_fft=True)

        @simple_in_memory_settings
        class Document(STFT):
            pass

        signal = SineSynthesizer(SR11025()).synthesize(Seconds(2))

        bio = BytesIO()
        filename = 'test.wav'
        with zipfile.ZipFile(bio, mode='w') as zf:
            zf.writestr(filename, signal.encode().read())
        bio.seek(0)

        zip_wrapper = featureflow.iter_zip(bio).next()
        _id = Document.process(meta=zip_wrapper)
        doc = Document(_id)
        self.assertEqual(2, doc.ogg.duration_seconds)



import csv

import numpy as np

from predownload import PreDownload
from util import ensure_local_file
from zounds.soundfile import AudioMetaData
from zounds.timeseries import SR44100, AudioSamples


class MusicNet(object):
    """
    Provides access to the audio and high-level metadata from MusicNet.  More
    info can be found here:
    https://homes.cs.washington.edu/~thickstn/musicnet.html
    """
    def __init__(self, path):
        super(MusicNet, self).__init__()
        self.path = path
        self._url = \
            'https://homes.cs.washington.edu/~thickstn/media/musicnet.npz'
        self._metadata = \
            'https://homes.cs.washington.edu/~thickstn/media/musicnet_metadata.csv'
        self._samplerate = SR44100()

    def __iter__(self):
        local_data = ensure_local_file(self._url, self.path)
        local_metadata = ensure_local_file(self._metadata, self.path)

        metadata = dict()
        with open(local_metadata, 'rb') as f:
            reader = csv.DictReader(f)
            for row in reader:
                metadata[row['id']] = row

        with open(local_data, 'rb') as f:
            data = np.load(f)
            for k, v in data.iteritems():
                _id = k
                samples, labels = v
                samples = AudioSamples(samples, self._samplerate)
                meta = metadata[_id]
                url = \
                    'https://homes.cs.washington.edu/~thickstn/media/{_id}'\
                        .format(**locals())
                meta['web_url'] = \
                    'https://homes.cs.washington.edu/~thickstn/musicnet.html'
                yield AudioMetaData(
                    uri=PreDownload(samples.encode().read(), url),
                    samplerate=int(SR44100()),
                    **meta)



import warnings
import featureflow as ff
from preprocess import Preprocessor, PreprocessResult, Op


class PyTorchPreprocessResult(PreprocessResult):
    def __init__(self, data, op, inversion_data=None, inverse=None, name=None):
        super(PyTorchPreprocessResult, self).__init__(
            data, op, inversion_data, inverse, name)

    def __getstate__(self):
        forward_func = self.op._func
        inv_data_func = self.inversion_data._func
        backward_func = self.inverse._func
        network_params = self.op.network.state_dict()
        weights = dict(
            ((k, v.cpu().numpy()) for k, v in network_params.iteritems()))
        cls = self.op.network.__class__
        name = self.name

        kwargs = self.op.kwargs
        del kwargs['network']

        return dict(
            forward_func=forward_func,
            op_kwargs=kwargs,
            inv_data_func=inv_data_func,
            backward_func=backward_func,
            weights=weights,
            name=name,
            cls=cls)

    def __setstate__(self, state):
        import torch
        restored_weights = dict(
            ((k, torch.from_numpy(v).cuda())
             for k, v in state['weights'].iteritems()))

        network = state['cls']()
        network.load_state_dict(restored_weights)
        network.cuda()
        network.eval()
        self.op = Op(
            state['forward_func'], network=network, **state['op_kwargs'])
        self.inversion_data = Op(state['inv_data_func'],
                                 network=network)
        self.inverse = Op(state['backward_func'])
        self.name = state['name']

    def for_storage(self):
        return PyTorchPreprocessResult(
            None,
            self.op,
            self.inversion_data,
            self.inverse,
            self.name)


class PyTorchNetwork(Preprocessor):
    def __init__(self, trainer=None, post_training_func=None, needs=None):

        super(PyTorchNetwork, self).__init__(needs=needs)
        self.trainer = trainer
        self.post_training_func = post_training_func or (lambda x: x)
        self._cache = dict()

    def _forward_func(self):
        def x(d, network=None):
            from zounds.core import ArrayWithUnits, IdentityDimension
            from zounds.learn import apply_network

            result = apply_network(network, d, chunksize=128)
            try:
                return ArrayWithUnits(
                    result, [d.dimensions[0], IdentityDimension()])
            except AttributeError:
                return result
            except ValueError:
                # the number of dimensions has likely changed
                return result

        return x

    def _backward_func(self):
        def x(_):
            raise NotImplementedError()

        return x

    def _enqueue(self, data, pusher):
        k = self._dependency_name(pusher)
        self._cache[k] = data

    def _dequeue(self):
        if not self._finalized:
            raise ff.NotEnoughData()
        return self._cache

    def _process(self, data):
        data = self._extract_data(data)

        trained_network = self.trainer.train(data)

        try:
            forward_func = self._forward_func()
            x = self.post_training_func(data['data'])
            processed_data = forward_func(x, network=trained_network)
        except RuntimeError as e:
            processed_data = None
            # the dataset may be too large to fit onto the GPU all at once
            warnings.warn(e.message)

        op = self.transform(network=trained_network)
        inv_data = self.inversion_data()
        inv = self.inverse_transform()

        yield PyTorchPreprocessResult(
            processed_data,
            op,
            inversion_data=inv_data,
            inverse=inv,
            name='PyTorchNetwork')


class PyTorchGan(PyTorchNetwork):
    def __init__(self, apply_network='generator', trainer=None, needs=None):
        super(PyTorchGan, self).__init__(trainer=trainer, needs=needs)

        if apply_network not in ('generator', 'discriminator'):
            raise ValueError(
                'apply_network must be one of (generator, discriminator)')

        self.apply_network = apply_network
        self._cache = None

    def _forward_func(self):
        def x(d, network=None, apply_network=None):
            import torch
            from torch.autograd import Variable
            import numpy as np
            from zounds.core import ArrayWithUnits, IdentityDimension

            if apply_network == 'generator':
                n = network.generator
            else:
                n = network.discriminator

            chunks = []
            batch_size = 128

            for i in xrange(0, len(d), batch_size):
                tensor = torch.from_numpy(
                    d[i:i + batch_size].astype(np.float32))
                gpu = tensor.cuda()
                v = Variable(gpu)
                chunks.append(n(v).data.cpu().numpy())

            result = np.concatenate(chunks)

            try:
                return ArrayWithUnits(
                    result, d.dimensions[:-1] + (IdentityDimension(),))
            except AttributeError:
                return result
            except ValueError:
                # the number of dimensions has likely changed
                return result

        return x

    def _backward_func(self):
        def x(_):
            raise NotImplementedError()

        return x

    def _enqueue(self, data, pusher):
        self._cache = data

    def _process(self, data):
        data = self._extract_data(data)

        network = self.trainer.train(data)
        network.eval()

        try:
            # note that the processed data passed on to the next step in the
            # training pipeline will be the labels output by the discriminator
            forward_func = self._forward_func()
            processed_data = forward_func(
                data, network=network, apply_network='discriminator')
        except RuntimeError as e:
            processed_data = None
            # the dataset may be too large to fit onto the GPU all at once
            warnings.warn(e.message)

        op = self.transform(network=network, apply_network=self.apply_network)
        print op.network
        print op.apply_network
        inv_data = self.inversion_data()
        inv = self.inverse_transform()

        yield PyTorchPreprocessResult(
            processed_data,
            op,
            inversion_data=inv_data,
            inverse=inv,
            name='PyTorchGan')


class PyTorchAutoEncoder(PyTorchNetwork):
    def __init__(self, trainer=None, needs=None):
        super(PyTorchAutoEncoder, self).__init__(trainer=trainer, needs=needs)
        self._cache = None

    def _forward_func(self):
        def x(d, network=None):
            import torch
            from torch.autograd import Variable
            import numpy as np
            from zounds.core import ArrayWithUnits, IdentityDimension

            chunks = []
            batch_size = 128

            for i in xrange(0, len(d), batch_size):
                tensor = torch.from_numpy(
                    d[i:i + batch_size].astype(np.float32))
                gpu = tensor.cuda()
                v = Variable(gpu)
                chunks.append(network.encoder(v).data.cpu().numpy())

            encoded = np.concatenate(chunks)

            try:
                extra_dims = (IdentityDimension(),) * (encoded.ndim - 1)
                return ArrayWithUnits(
                    encoded, d.dimensions[:1] + extra_dims)
            except AttributeError:
                return encoded

        return x

    def _backward_func(self):
        def x(d, network=None):
            import torch
            from torch.autograd import Variable
            import numpy as np

            chunks = []
            batch_size = 128

            for i in xrange(0, len(d), batch_size):
                tensor = torch.from_numpy(d.astype(np.float32))
                gpu = tensor.cuda()
                v = Variable(gpu)
                chunks.append(network.decoder(v).data.cpu().numpy())

            decoded = np.concatenate(chunks)
            return decoded

        return x

    def _enqueue(self, data, pusher):
        self._cache = data

    def _process(self, data):
        data = self._extract_data(data)

        data = dict(data=data, labels=data)

        trained_network = self.trainer.train(data)
        trained_network.eval()

        processed_data = None
        inp = data['data']

        while processed_data is None:
            try:
                forward_func = self._forward_func()
                processed_data = forward_func(inp, network=trained_network)
            except RuntimeError as e:
                processed_data = None
                warnings.warn(e.message)
                # we've just experienced an out of memory exception.  Cut the
                # size of the input data in half, so that downstream nodes that
                # need some data to initialize themselves can do so
                inp = inp[:len(inp) // 64]
            except ValueError:
                break

        op = self.transform(network=trained_network)
        inv_data = self.inversion_data(network=trained_network)
        inv = self.inverse_transform()

        yield PyTorchPreprocessResult(
            processed_data,
            op,
            inversion_data=inv_data,
            inverse=inv,
            name='PyTorchAutoEncoder')



from __future__ import division

import featureflow as ff
import numpy as np
import unittest2

from preprocess import \
    UnitNorm, Binarize, PreprocessingPipeline, InstanceScaling
from pytorch_model import PyTorchNetwork, PyTorchAutoEncoder, PyTorchGan
from supervised import SupervisedTrainer
from wgan import WassersteinGanTrainer
from random_samples import ShuffledSamples
from zounds.core import ArrayWithUnits, IdentityDimension
from zounds.spectral import LinearScale, FrequencyBand, FrequencyDimension
from zounds.timeseries import Seconds, TimeDimension
from zounds.util import simple_in_memory_settings

try:
    import torch
    from torch import nn
    from torch.optim import SGD, Adam


    class SupervisedNetwork(nn.Module):
        def __init__(self):
            super(SupervisedNetwork, self).__init__()
            self.visible = nn.Linear(2, 64, bias=False)
            self.t1 = nn.Sigmoid()
            self.hidden = nn.Linear(64, 1, bias=False)
            self.t2 = nn.Sigmoid()

        def forward(self, inp):
            x = self.visible(inp)
            x = self.t1(x)
            x = self.hidden(x)
            x = self.t2(x)
            return x


    class AutoEncoder(nn.Module):
        def __init__(self):
            super(AutoEncoder, self).__init__()

            self.encoder = nn.Sequential(
                nn.Linear(3, 2, bias=False),
                nn.Sigmoid())

            self.decoder = nn.Sequential(
                nn.Linear(2, 3, bias=False),
                nn.Sigmoid())

        def forward(self, inp):
            x = self.encoder(inp)
            x = self.decoder(x)
            return x


    class GanGenerator(nn.Module):
        def __init__(self):
            super(GanGenerator, self).__init__()
            self.linear = nn.Linear(2, 4)
            self.tanh = nn.Tanh()

        def forward(self, inp):
            x = self.linear(inp)
            x = self.tanh(x)
            return x


    class GanDiscriminator(nn.Module):
        def __init__(self):
            super(GanDiscriminator, self).__init__()
            self.linear = nn.Linear(4, 1)
            self.sigmoid = nn.Sigmoid()

        def forward(self, inp):
            x = self.linear(inp)
            x = self.sigmoid(x)
            return x

    class GanPair(nn.Module):
        def __init__(self):
            super(GanPair, self).__init__()
            self.generator = GanGenerator()
            self.discriminator = GanDiscriminator()

        def forward(self, x):
            raise NotImplementedError()

except ImportError:
    torch = None


class PyTorchModelTests(unittest2.TestCase):
    def setUp(self):
        if torch is None:
            self.skipTest('pytorch is not available')

    def test_can_maintain_array_dimensions_with_supervised_learning(self):
        trainer = SupervisedTrainer(
            model=SupervisedNetwork(),
            loss=nn.BCELoss(),
            optimizer=lambda model: SGD(model.parameters(), lr=0.2),
            epochs=1,
            batch_size=64)

        @simple_in_memory_settings
        class Pipeline(ff.BaseModel):
            inp = ff.PickleFeature(
                ff.IteratorNode,
                store=False)

            samples = ff.PickleFeature(
                ShuffledSamples,
                nsamples=500,
                multiplexed=True,
                needs=inp,
                store=False)

            unitnorm = ff.PickleFeature(
                UnitNorm,
                needs=samples.aspect('data'),
                store=False)

            hard_labels = ff.PickleFeature(
                Binarize,
                needs=samples.aspect('labels'),
                store=False)

            network = ff.PickleFeature(
                PyTorchNetwork,
                trainer=trainer,
                needs=dict(data=unitnorm, labels=hard_labels),
                store=False)

            pipeline = ff.PickleFeature(
                PreprocessingPipeline,
                needs=(unitnorm, network),
                store=True)

        # Produce some random points on the unit circle
        samples = np.random.random_sample((1000, 2))
        samples /= np.linalg.norm(samples, axis=1, keepdims=True)

        # a line extending from the origin to (1, 1)
        origin = np.array([0, 0])
        unit = np.array([1, 1])

        # which side of the plane is each sample on?
        labels = np.sign(np.cross(unit - origin, origin - samples))
        labels[labels < 0] = 0

        # scale each sample randomly, forcing the pipeline to normalize data
        factors = np.random.randint(1, 1000, (len(samples), 1))
        scaled_samples = samples * factors
        scaled_samples = scaled_samples

        # fuzz the labels, forcing the pipeline to binarize these (i.e., force
        # them to be 0 or 1)
        fuzzed_labels = labels + np.random.normal(0, 0.1, labels.shape)
        fuzzed_labels = fuzzed_labels[..., None]

        def gen(chunksize, s, l):
            for i in xrange(0, len(s), chunksize):
                sl = slice(i, i + chunksize)
                yield dict(data=s[sl], labels=l[sl])

        _id = Pipeline.process(inp=gen(100, scaled_samples, fuzzed_labels))
        pipe = Pipeline(_id)

        # produce some new samples
        new_samples = np.random.random_sample((1000, 2))
        new_samples /= np.linalg.norm(samples, axis=1, keepdims=True)

        # scale each example randomly, so the pipeline must give it unit norm
        # to arrive at the correct answer
        new_factors = np.random.randint(1, 1000, (len(samples), 1))
        new_scaled_samples = new_factors * new_samples

        arr = ArrayWithUnits(
            new_scaled_samples,
            dimensions=[
                TimeDimension(Seconds(1)),
                FrequencyDimension(LinearScale(FrequencyBand(100, 1000), 2))
            ])

        result = pipe.pipeline.transform(arr)
        self.assertIsInstance(result.data, ArrayWithUnits)
        self.assertIsInstance(result.data.dimensions[0], TimeDimension)

    def test_can_perform_supervised_learning(self):
        """
        Create and exercise a learning pipeline that learn to classify
        2d points as being on one side or the other of a plane from (0, 0) to
        (1, 1)
        """

        trainer = SupervisedTrainer(
            model=SupervisedNetwork(),
            loss=nn.BCELoss(),
            optimizer=lambda model: SGD(model.parameters(), lr=0.2),
            epochs=100,
            batch_size=64)

        @simple_in_memory_settings
        class Pipeline(ff.BaseModel):
            inp = ff.PickleFeature(
                ff.IteratorNode,
                store=False)

            samples = ff.PickleFeature(
                ShuffledSamples,
                nsamples=500,
                multiplexed=True,
                needs=inp,
                store=False)

            unitnorm = ff.PickleFeature(
                UnitNorm,
                needs=samples.aspect('data'),
                store=False)

            hard_labels = ff.PickleFeature(
                Binarize,
                needs=samples.aspect('labels'),
                store=False)

            network = ff.PickleFeature(
                PyTorchNetwork,
                trainer=trainer,
                needs=dict(data=unitnorm, labels=hard_labels),
                store=False)

            pipeline = ff.PickleFeature(
                PreprocessingPipeline,
                needs=(unitnorm, network),
                store=True)

        # Produce some random points on the unit circle
        samples = np.random.random_sample((1000, 2))
        samples /= np.linalg.norm(samples, axis=1, keepdims=True)

        # a line extending from the origin to (1, 1)
        origin = np.array([0, 0])
        unit = np.array([1, 1])

        # which side of the plane is each sample on?
        labels = np.sign(np.cross(unit - origin, origin - samples))
        labels[labels < 0] = 0

        # scale each sample randomly, forcing the pipeline to normalize data
        factors = np.random.randint(1, 1000, (len(samples), 1))
        scaled_samples = samples * factors
        scaled_samples = scaled_samples

        # fuzz the labels, forcing the pipeline to binarize these (i.e., force
        # them to be 0 or 1)
        fuzzed_labels = labels + np.random.normal(0, 0.1, labels.shape)
        fuzzed_labels = fuzzed_labels[..., None]

        def gen(chunksize, s, l):
            for i in xrange(0, len(s), chunksize):
                sl = slice(i, i + chunksize)
                yield dict(data=s[sl], labels=l[sl])

        _id = Pipeline.process(inp=gen(100, scaled_samples, fuzzed_labels))
        pipe = Pipeline(_id)

        # produce some new samples
        new_samples = np.random.random_sample((1000, 2))
        new_samples /= np.linalg.norm(samples, axis=1, keepdims=True)

        # which side of the plane is each sample on?
        new_labels = np.sign(np.cross(unit - origin, origin - new_samples))
        new_labels[new_labels < 0] = 0

        # scale each example randomly, so the pipeline must give it unit norm
        # to arrive at the correct answer
        new_factors = np.random.randint(1, 1000, (len(samples), 1))
        new_scaled_samples = new_factors * new_samples

        result = pipe.pipeline.transform(new_scaled_samples)

        # reshape the data, and normalize to 0 or 1
        result = np.round(result.data.squeeze())

        # compute the number of labels that are incorrect
        difference = np.logical_xor(result, new_labels).sum()
        percent_error = difference / len(result)

        self.assertLess(percent_error, 0.05)

    def test_maintains_array_with_units_dimensions(self):
        trainer = SupervisedTrainer(
            AutoEncoder(),
            loss=nn.MSELoss(),
            optimizer=lambda model: SGD(model.parameters(), lr=0.1),
            epochs=2,
            batch_size=64)

        @simple_in_memory_settings
        class Pipeline(ff.BaseModel):
            inp = ff.PickleFeature(
                ff.IteratorNode,
                store=False)

            samples = ff.PickleFeature(
                ShuffledSamples,
                nsamples=500,
                needs=inp,
                store=False)

            unitnorm = ff.PickleFeature(
                UnitNorm,
                needs=samples,
                store=False)

            network = ff.PickleFeature(
                PyTorchAutoEncoder,
                trainer=trainer,
                needs=unitnorm,
                store=False)

            pipeline = ff.PickleFeature(
                PreprocessingPipeline,
                needs=(unitnorm, network),
                store=True)

        training = np.random.random_sample((1000, 3))

        def gen(chunksize, s):
            for i in xrange(0, len(s), chunksize):
                yield s[i: i + chunksize]

        _id = Pipeline.process(inp=gen(100, training))
        pipe = Pipeline(_id)

        test = ArrayWithUnits(
            np.random.random_sample((10, 3)),
            dimensions=[
                TimeDimension(Seconds(1)),
                FrequencyDimension(LinearScale(FrequencyBand(100, 1000), 3))
            ])
        result = pipe.pipeline.transform(test)
        self.assertEqual((10, 2), result.data.shape)
        self.assertIsInstance(result.data, ArrayWithUnits)
        self.assertIsInstance(result.data.dimensions[0], TimeDimension)
        self.assertIsInstance(result.data.dimensions[1], IdentityDimension)

        inverted = result.inverse_transform()
        self.assertEqual((10, 3), inverted.shape)
        self.assertIsInstance(inverted, ArrayWithUnits)
        self.assertIsInstance(inverted.dimensions[0], TimeDimension)
        self.assertIsInstance(inverted.dimensions[1], FrequencyDimension)

    def test_can_perform_unsupervised_learning_autoencoder(self):

        trainer = SupervisedTrainer(
            AutoEncoder(),
            loss=nn.MSELoss(),
            optimizer=lambda model: SGD(model.parameters(), lr=0.1),
            epochs=10,
            batch_size=64)

        @simple_in_memory_settings
        class Pipeline(ff.BaseModel):
            inp = ff.PickleFeature(
                ff.IteratorNode,
                store=False)

            samples = ff.PickleFeature(
                ShuffledSamples,
                nsamples=500,
                needs=inp,
                store=False)

            unitnorm = ff.PickleFeature(
                UnitNorm,
                needs=samples,
                store=False)

            network = ff.PickleFeature(
                PyTorchAutoEncoder,
                trainer=trainer,
                needs=unitnorm,
                store=False)

            pipeline = ff.PickleFeature(
                PreprocessingPipeline,
                needs=(unitnorm, network),
                store=True)

        training = np.random.random_sample((1000, 3))

        def gen(chunksize, s):
            for i in xrange(0, len(s), chunksize):
                yield s[i: i + chunksize]

        _id = Pipeline.process(inp=gen(100, training))
        pipe = Pipeline(_id)

        test = np.random.random_sample((10, 3))
        result = pipe.pipeline.transform(test)
        self.assertEqual((10, 2), result.data.shape)

        inverted = result.inverse_transform()
        self.assertEqual((10, 3), inverted.shape)

    def test_can_train_gan(self):

        trainer = WassersteinGanTrainer(
            GanPair(),
            latent_dimension=(2,),
            n_critic_iterations=5,
            epochs=10,
            batch_size=64)

        @simple_in_memory_settings
        class Pipeline(ff.BaseModel):
            inp = ff.PickleFeature(
                ff.IteratorNode,
                store=False)

            samples = ff.PickleFeature(
                ShuffledSamples,
                nsamples=500,
                needs=inp,
                store=False)

            scaled = ff.PickleFeature(
                InstanceScaling,
                needs=samples,
                store=False)

            network = ff.PickleFeature(
                PyTorchGan,
                apply_network='generator',
                trainer=trainer,
                needs=scaled,
                store=False)

            pipeline = ff.PickleFeature(
                PreprocessingPipeline,
                needs=(scaled, network),
                store=True)

        training_data = np.random.normal(0, 1, (1000, 4))

        def gen(chunksize, s):
            for i in xrange(0, len(s), chunksize):
                yield s[i: i + chunksize]

        _id = Pipeline.process(inp=gen(100, training_data))
        pipe = Pipeline(_id)

        noise = np.random.normal(0, 1, (10, 2))
        result = pipe.pipeline.transform(noise)
        self.assertEqual((10, 4), result.data.shape)



import numpy as np


class Hours(np.timedelta64):
    """
    Convenience class for creating a duration in hours

    Args:
        hours (int): duration in hours

    Examples:
        >>> from zounds import Hours
        >>> hours = Hours(3)
        >>> hours
        numpy.timedelta(3, 'h')
    """
    def __new__(cls, hours):
        return np.timedelta64(hours, 'h')


class Minutes(np.timedelta64):
    """
    Convenience class for creating a duration in minutes

    Args:
        minutes (int): duration in minutes

    Examples:
        >>> from zounds import Minutes
        >>> minutes = Minutes(3)
        >>> minutes
        numpy.timedelta(3, 'm')
    """
    def __new__(cls, minutes):
        return np.timedelta64(minutes, 'm')


class Seconds(np.timedelta64):
    """
    Convenience class for creating a duration in seconds

    Args:
        seconds (int): duration in seconds

    Examples:
        >>> from zounds import Seconds
        >>> seconds = Seconds(3)
        >>> seconds
        numpy.timedelta(3, 's')
    """
    def __new__(cls, seconds):
        return np.timedelta64(seconds, 's')


class Milliseconds(np.timedelta64):
    """
    Convenience class for creating a duration in milliseconds

    Args:
        milliseconds (int): duration in milliseconds

    Examples:
        >>> from zounds import Milliseconds
        >>> ms = Milliseconds(3)
        >>> ms
        numpy.timedelta(3, 'ms')
    """
    def __new__(cls, milliseconds):
        return np.timedelta64(milliseconds, 'ms')


class Microseconds(np.timedelta64):
    """
    Convenience class for creating a duration in microseconds

    Args:
        microseconds (int): duration in microseconds

    Examples:
        >>> from zounds import Microseconds
        >>> us = Microseconds(3)
        >>> us
        numpy.timedelta(3, 'us')
    """
    def __new__(cls, microseconds):
        return np.timedelta64(microseconds, 'us')


class Nanoseconds(np.timedelta64):
    """
    Convenience class for creating a duration in nanoseconds

    Args:
        nanoseconds (int): duration in nanoseconds

    Examples:
        >>> from zounds import Nanoseconds
        >>> ns = Nanoseconds(3)
        >>> ns
        numpy.timedelta(3, 'ns')
    """
    def __new__(cls, nanoseconds):
        return np.timedelta64(nanoseconds, 'ns')


class Picoseconds(np.timedelta64):
    """
    Convenience class for creating a duration in picoseconds

    Args:
        picoseconds (int): duration in picoseconds

    Examples:
        >>> from zounds import Picoseconds
        >>> ps = Picoseconds(3)
        >>> ps
        numpy.timedelta(3, 'ps')
    """
    def __new__(cls, picoseconds):
        return np.timedelta64(picoseconds, 'ps')



from random import random

import featureflow as ff
import numpy as np
import unittest2

from duration import Milliseconds
from timeseries import TimeSlice
from variablerate import VariableRateTimeSeries, VariableRateTimeSeriesFeature
from zounds.basic import Pooled, stft
from zounds.segment import TimeSliceFeature
from zounds.synthesize import NoiseSynthesizer
from zounds.timeseries import Picoseconds, Seconds, SR44100
from zounds.util import simple_in_memory_settings


class VariableRateTimeSeriesFeatureTests(unittest2.TestCase):

    def test_can_encode_and_decode_variable_rate_time_Series(self):

        class TimestampEmitter(ff.Node):
            def __init__(self, needs=None):
                super(TimestampEmitter, self).__init__(needs=needs)
                self.pos = Picoseconds(0)

            def _process(self, data):
                td = data.dimensions[0]
                frequency = td.frequency
                timestamps = [self.pos + (i * frequency)
                        for i, d in enumerate(data)
                        if random() > 0.9]
                slices = TimeSlice.slices(timestamps)
                yield VariableRateTimeSeries(
                    (ts, np.zeros(0)) for ts in slices)
                self.pos += frequency * len(data)

        graph = stft(store_fft=True)

        @simple_in_memory_settings
        class Document(graph):
            slices = TimeSliceFeature(
                    TimestampEmitter,
                    needs=graph.fft,
                    store=True)

            pooled = VariableRateTimeSeriesFeature(
                    Pooled,
                    op=np.max,
                    axis=0,
                    needs=(slices, graph.fft),
                    store=False)

        signal = NoiseSynthesizer(SR44100())\
            .synthesize(Seconds(10))\
            .encode()
        _id = Document.process(meta=signal)
        doc = Document(_id)
        self.assertIsInstance(doc.pooled, VariableRateTimeSeries)
        self.assertEqual(doc.fft.shape[1], doc.pooled.slicedata.shape[1])


class VariableRateTimeSeriesTests(unittest2.TestCase):

    def test_can_concatenate_variable_rate_time_series(self):
        ts1 = VariableRateTimeSeries((
            (TimeSlice(start=Seconds(1), duration=Seconds(1)), np.zeros(10)),
            (TimeSlice(start=Seconds(2), duration=Seconds(1)), np.zeros(10)),
        ))
        ts2 = VariableRateTimeSeries((
            (TimeSlice(start=Seconds(1), duration=Seconds(1)), np.zeros(10)),
            (TimeSlice(start=Seconds(2), duration=Seconds(1)), np.zeros(10)),
        ))
        ts3 = ts1.concat(ts2)
        self.assertEqual(4, len(ts3))
        self.assertEqual((4, 10), ts3.slicedata.shape)

    def test_concat_fails_when_data_shape_is_mismatched(self):
        ts1 = VariableRateTimeSeries((
            (TimeSlice(start=Seconds(1), duration=Seconds(1)), np.zeros(10)),
            (TimeSlice(start=Seconds(2), duration=Seconds(1)), np.zeros(10)),
        ))
        ts2 = VariableRateTimeSeries((
            (TimeSlice(start=Seconds(1), duration=Seconds(1)), np.zeros(11)),
            (TimeSlice(start=Seconds(2), duration=Seconds(1)), np.zeros(11)),
        ))
        self.assertRaises(ValueError, lambda: ts1.concat(ts2))

    def test_can_create_instance_with_no_slice_data(self):
        ts = VariableRateTimeSeries((
            (TimeSlice(start=Seconds(1), duration=Seconds(1)), np.zeros(0)),
            (TimeSlice(start=Seconds(2), duration=Seconds(1)), np.zeros(0)),
        ))
        self.assertEqual(2, len(ts))
        self.assertEqual((2, 0), ts.slicedata.shape)

    def test_can_slice_time_series_with_time_slice(self):
        ts = VariableRateTimeSeries((
            (TimeSlice(start=Seconds(0), duration=Seconds(1)), np.zeros(10)),
            (TimeSlice(start=Seconds(1), duration=Seconds(2)), np.zeros(10)),
            (TimeSlice(start=Seconds(3), duration=Seconds(1)), np.zeros(10))
        ))
        sliced = ts[TimeSlice(start=Seconds(1), duration=Seconds(2))]
        self.assertIsInstance(sliced, VariableRateTimeSeries)
        self.assertEqual(1, len(sliced))

    def test_can_slice_time_series_with_open_ended_time_slice(self):
        ts = VariableRateTimeSeries((
            (TimeSlice(start=Seconds(0), duration=Seconds(1)), np.zeros(10)),
            (TimeSlice(start=Seconds(1), duration=Seconds(2)), np.zeros(10)),
            (TimeSlice(start=Seconds(3), duration=Seconds(1)), np.zeros(10))
        ))
        sliced = ts[TimeSlice(start=Seconds(1))]
        self.assertIsInstance(sliced, VariableRateTimeSeries)
        self.assertEqual(2, len(sliced))

    def test_can_index_time_series_with_integer_index(self):
        ts = VariableRateTimeSeries((
            (TimeSlice(start=Seconds(0), duration=Seconds(1)), np.zeros(10)),
            (TimeSlice(start=Seconds(1), duration=Seconds(2)), np.zeros(10)),
            (TimeSlice(start=Seconds(3), duration=Seconds(1)), np.zeros(10))
        ))
        sliced = ts[1]
        self.assertIsInstance(sliced, np.record)
        timeslice, data = sliced
        self.assertEqual(
                TimeSlice(start=Seconds(1), duration=Seconds(2)), timeslice)
        self.assertIsInstance(data, np.ndarray)
        self.assertEqual((10,), data.shape)

    def test_can_slice_time_series_with_integer_indices(self):
        ts = VariableRateTimeSeries((
            (TimeSlice(start=Seconds(0), duration=Seconds(1)), np.zeros(10)),
            (TimeSlice(start=Seconds(1), duration=Seconds(2)), np.zeros(10)),
            (TimeSlice(start=Seconds(3), duration=Seconds(1)), np.zeros(10))
        ))
        sliced = ts[1:]
        self.assertIsInstance(sliced, VariableRateTimeSeries)
        self.assertEqual(2, len(sliced))

    def test_get_index_error_when_using_out_of_range_int_index(self):
        ts = VariableRateTimeSeries((
            (TimeSlice(start=Seconds(0), duration=Seconds(1)), np.zeros(10)),
            (TimeSlice(start=Seconds(1), duration=Seconds(2)), np.zeros(10)),
            (TimeSlice(start=Seconds(3), duration=Seconds(1)), np.zeros(10))
        ))
        self.assertRaises(IndexError, lambda: ts[5])

    def test_get_empty_time_series_when_using_out_of_range_time_slice(self):
        ts = VariableRateTimeSeries((
            (TimeSlice(start=Seconds(0), duration=Seconds(1)), np.zeros(10)),
            (TimeSlice(start=Seconds(1), duration=Seconds(2)), np.zeros(10)),
            (TimeSlice(start=Seconds(3), duration=Seconds(1)), np.zeros(10))
        ))
        sliced = ts[TimeSlice(start=Seconds(10), duration=Seconds(1))]
        self.assertIsInstance(sliced, VariableRateTimeSeries)
        self.assertEqual(0, len(sliced))

    def test_time_slice_spanning_less_than_one_example_returns_one_example(
            self):
        ts = VariableRateTimeSeries((
            (TimeSlice(start=Seconds(0), duration=Seconds(1)), np.zeros(10)),
            (TimeSlice(start=Seconds(1), duration=Seconds(2)), np.zeros(10)),
            (TimeSlice(start=Seconds(3), duration=Seconds(1)), np.zeros(10))
        ))
        timeslice = TimeSlice(
                start=Milliseconds(500), duration=Milliseconds(100))
        sliced = ts[timeslice]
        self.assertIsInstance(sliced, VariableRateTimeSeries)
        self.assertEqual(1, len(sliced))

    def test_time_slice_spanning_multiple_examples_returns_all_examples(self):
        ts = VariableRateTimeSeries((
            (TimeSlice(start=Seconds(0), duration=Seconds(1)), np.zeros(10)),
            (TimeSlice(start=Seconds(1), duration=Seconds(2)), np.zeros(10)),
            (TimeSlice(start=Seconds(3), duration=Seconds(1)), np.zeros(10))
        ))
        timeslice = TimeSlice(start=Milliseconds(500), duration=Seconds(1))
        sliced = ts[timeslice]
        self.assertIsInstance(sliced, VariableRateTimeSeries)
        self.assertEqual(2, len(sliced))

    def test_can_get_entire_time_series_with_empty_slice(self):
        ts = VariableRateTimeSeries((
            (TimeSlice(start=Seconds(0), duration=Seconds(1)), np.zeros(10)),
            (TimeSlice(start=Seconds(1), duration=Seconds(2)), np.zeros(10)),
            (TimeSlice(start=Seconds(3), duration=Seconds(1)), np.zeros(10))
        ))
        sliced = ts[:]
        self.assertIsInstance(sliced, VariableRateTimeSeries)
        self.assertEqual(3, len(sliced))

    def test_sorts_input(self):
        ts = VariableRateTimeSeries((
            (TimeSlice(start=Seconds(3), duration=Seconds(1)), np.zeros(10)),
            (TimeSlice(start=Seconds(0), duration=Seconds(1)), np.zeros(10)),
            (TimeSlice(start=Seconds(1), duration=Seconds(2)), np.zeros(10))
        ))
        timeslice, data = ts[0]
        self.assertEqual(
                TimeSlice(start=Seconds(0), duration=Seconds(1)), timeslice)

    def raises_if_data_is_of_variable_size(self):
        data = ((
            (TimeSlice(start=Seconds(3), duration=Seconds(1)), np.zeros(11)),
            (TimeSlice(start=Seconds(0), duration=Seconds(1)), np.zeros(10)),
            (TimeSlice(start=Seconds(1), duration=Seconds(2)), np.zeros(10))
        ))
        self.assertRaises(ValueError, VariableRateTimeSeries(data))

    def test_span(self):
        ts = VariableRateTimeSeries((
            (TimeSlice(start=Seconds(0), duration=Seconds(1)), np.zeros(10)),
            (TimeSlice(start=Seconds(1), duration=Seconds(2)), np.zeros(10)),
            (TimeSlice(start=Seconds(3), duration=Seconds(1)), np.zeros(10))
        ))
        self.assertEqual(
                TimeSlice(start=Seconds(0), duration=Seconds(4)), ts.span)

    def test_span_empty(self):
        ts = VariableRateTimeSeries(())
        self.assertEqual(
                TimeSlice(start=Seconds(0), duration=Seconds(0)), ts.span)

    def test_end(self):
        ts = VariableRateTimeSeries((
            (TimeSlice(start=Seconds(0), duration=Seconds(1)), np.zeros(10)),
            (TimeSlice(start=Seconds(1), duration=Seconds(2)), np.zeros(10)),
            (TimeSlice(start=Seconds(3), duration=Seconds(1)), np.zeros(10))
        ))
        self.assertEqual(Seconds(4), ts.end)

    def test_end_empty(self):
        ts = VariableRateTimeSeries(())
        self.assertEqual(Seconds(0), ts.end)

